/***************************************************************************
 *cr                                                                       
 *cr            (C) Copyright 1995 The Board of Trustees of the           
 *cr                        University of Illinois                       
 *cr                         All Rights Reserved                        
 *cr                                                                   
 ***************************************************************************/

/***************************************************************************
 * RCS INFORMATION:
 *
 *	$RCSfile: DrawMolItem.C,v $
 *	$Author: ulrich $	$Locker:  $		$State: Exp $
 *	$Revision: 1.55 $	$Date: 1997/03/21 22:14:53 $
 *
 ***************************************************************************
 * DESCRIPTION:
 *
 * Child Displayable component of a molecule; this is responsible for doing
 * the actual drawing of a molecule.  It contains an atom color, atom
 * selection, and atom representation object to specify how this component
 * should look.
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>		// for systems()
#include <unistd.h>
#include <sys/types.h>
#include <math.h>

#include "DrawMolItem.h"
#include "DrawMolecule.h"
extern MoleculeList *moleculeList;
#include "DispCmds.h"
#include "PopupMenu.h"
#include "Inform.h"
#include "Scene.h"
#include "CoorPDB.h"		// for write_pdb_record

#ifdef VMDALPHA_SHAPE
// This is not yet available outside of the Theoretical Biophysics
// Group and is still under development.
#include "AlphaShape.h"
#endif

#ifdef VMDSURF
// this is an interface to the SURF program
#include "Surf.h"
#endif

// draw commands used by this object to indicate points to pick
static DispCmdPickPoint pickPoint;
static DispCmdPickPointIndex pickPointIndex;

static int drawmolitem_tube_radius_hack = 0;

//////////////////////////  constructor  

DrawMolItem::DrawMolItem(DrawMolecule *dm, AtomColor *ac, AtomRep *ar,
	AtomSel *as) 
	: Displayable3D(MULT, dm->name, dm, dm->nAtoms/8 + 2),
	  cmdTextX("."),cmdCommentX(".") {

  MSGDEBUG(1,"Creating new molecule representation, for parent molecule:\n");
  MSGDEBUG(1," ==> '" << dm->name << "'" << sendmsg);

  // save data and pointers to drawing method objects
  mol = dm;
  atomColor = ac;
  atomRep = ar;
  atomSel = as;
  
  // create some useful display commands
  dataCmd = new DispCmdData(3 * mol->nAtoms, NULL);

  // signal that we need a complete refresh next prepare
  needRegenerate = FORCE_REGEN;
  needNewFrame = TRUE;

  // register with the pick list, to indicate we have objects that may
  // be selected with the mouse
  register_with_picklist(this->origScene);
}

//////////////////////////  destructor  
DrawMolItem::~DrawMolItem(void) {

  MSGDEBUG(2,"Deleting DrawMolItem '" << name << "' ..." << sendmsg);

  delete atomColor;
  delete atomRep;
  delete atomSel;

  delete dataCmd;
}


//////////////////////////////// private routines 

// regenerate the command list
void DrawMolItem::create_cmdlist(void) {
  float *framepos;
  int i;

  // Identify the representation number for output comments
  repNumber = -1;   // Nonsense default
  for (i = 0; i < mol->components(); i++) {
    if (this == mol->component(i))
      repNumber = i;
  }

  // do we need to recreate everything?
  if(needRegenerate) {
  
    MSGDEBUG(2, "Regenerating display list for molecule '"<< name<< "' ...");
    MSGDEBUG(2, sendmsg);

    // regenerate both data block and display commands
    needNewFrame = FALSE;
    reset_disp_list();

    // only put in commands if there is a current frame
    if(mol->frame() >= 0) {
      // get current position coordinates
      framepos = (mol->current())->pos;

      // first put in data command
      dataCmd->putdata(framepos, this);
      
      // now put in drawing commands, which index into the above block
      switch (atomRep->method()) {
      case AtomRep::LINES:     draw_lines(framepos); break;
      case AtomRep::BONDS:     draw_bonds(framepos); break;
      case AtomRep::CPK:       draw_cpk(framepos); break;
      case AtomRep::POINTS:    draw_points(framepos); break;
      case AtomRep::VDW:       draw_vdw(); break;
      case AtomRep::DOTTED:    draw_dotted(); break;

      // this function is in DrawMolItemSurface.C
      case AtomRep::TRACE:     draw_trace(framepos); place_picks(); break;
      case AtomRep::SOLVENT:   draw_dot_surface(framepos); 
	                       place_picks(); break;
      case AtomRep::LICORICE:  draw_licorice(framepos); break;
      case AtomRep::RIBBONS:   draw_ribbons(framepos); break;
      case AtomRep::TUBE:      draw_tube(framepos); break;
      case AtomRep::STRUCTURE: draw_structure(framepos); break;
      case AtomRep::HBONDS:    draw_hbonds(framepos); break;
	
#ifdef VMDALPHA_SHAPE
      case AtomRep::ALPHASHAPE: draw_alphaShape(framepos); break;
#endif
#ifdef VMDSURF
      // this function is in DrawMolItemSurface.C
    case AtomRep::SURF:
      draw_surface(framepos);
      break;
#endif
    case AtomRep::REPOFF: break; // don't do anything ... draws a blank object
    default:
        msgErr << "Illegal atom representation in DrawMolecule." << sendmsg;
    }
    }

  } else if(needNewFrame) {
    // just update the coordinates ... assume there is a current frame
    needNewFrame = FALSE;
    dataCmd->reputdata((mol->current())->pos, this);
  }
  // and, finally, I don't need a new regeneration
  needRegenerate = NO_REGEN;
}

/////////////////////////////////
// put down the pick points
void DrawMolItem::place_picks(void)
{
  for (int i=0; i < mol -> nAtoms; i++) {
    if (atomSel->on[i]) {
      pickPointIndex.putdata(3*i, i, this);
    }
  }
}


//////////////////////////////// protected virtual routines
// do action due to the fact that a color for the given ColorList for
// the specified category has changed
void DrawMolItem::do_color_changed(ColorList *collist, int ccat) {

  // check the category ... see if is for the current coloring method
  if(atomColor && atomColor->current_color_use(collist, ccat))
    change_color(atomColor);		// recalcs colors; signals need redraw
}


//////////////////////////////// public routines 

// return which representation this is, i.e. an index from 0 ... # reps - 1
int DrawMolItem::representation_index(void) {

  int totalreps = mol->components();
  for (int i=0; i < totalreps; i++)
    if (mol->component(i) == this)
      return i;

  // if here, error that should not occur
  msgErr << "Unknown molecular representation found." << sendmsg;
  return 0;
}


// common atom selection methods used in the following routine
static char *common_atom_selections[] = { "backbone", "protein", "nucleic",
 "waters", "not waters", "{not name \"[0-9]?H.*\"}"};
static char *common_atom_sel_names[] = { "backbone", "protein", "nucleic",
 "waters", "not waters", "heavy atoms"};

// Create a popup menu, with the given name; return NULL if none available
// subclasses should have the parent first create a menu, and then add items
// to the end of it.  The argument is the tag of the selected item; here, it is
// the selected atom.
PopupMenu *DrawMolItem::create_popup_menu(int atomIndex) {
  register int i;

  // special case: get menu from parent molecule as starting point
  PopupMenu *pm = mol->create_popup_menu(atomIndex);

  // if we have a specialized menu from an earlier class, add it as a submenu
  PopupMenu *pm2 = Displayable::create_popup_menu(atomIndex);

  // create or add menus as necessary
  if(pm && pm2) {
    pm->add_submenu(pm2, TRUE, FALSE);
  } else if(pm2) {
    pm = pm2;
  } else if(!pm) {
    pm = new PopupMenu(name);
  }

  //
  // add items to the menu
  //
  char cmdbuf[128];
  int repindex = representation_index();
  int molindex = mol->id();

  // add submenu to choose common atom selections
  PopupMenu *pmSel = new PopupMenu("Atom Selection");
  int sels = sizeof(common_atom_selections) / sizeof(char *);
  // first some things related to this atom
  sprintf(cmdbuf, "mol modselect %d %d same residue as index %d",
	  repindex, molindex, atomIndex);
  pmSel->add_item("Same Residue", cmdbuf, TRUE, FALSE);
  sprintf(cmdbuf, "mol modselect %d %d same chain as index %d",
	  repindex, molindex, atomIndex);
  pmSel->add_item("Same Chain", cmdbuf, TRUE, FALSE);
  sprintf(cmdbuf, "mol modselect %d %d same segname as index %d",
	  repindex, molindex, atomIndex);
  pmSel->add_item("Same Segname", cmdbuf, TRUE, FALSE);
  sprintf(cmdbuf, "mol modselect %d %d same fragment as index %d",
	  repindex, molindex, atomIndex);
  pmSel->add_item("Same Fragment", cmdbuf, TRUE, FALSE);

  for(i=0; i < sels; i++) {
    sprintf(cmdbuf, "mol modselect %d %d %s", repindex, molindex,
	    common_atom_selections[i]);
    pmSel->add_item(common_atom_sel_names[i], cmdbuf, TRUE, FALSE);
  }
  pm->add_submenu(pmSel, TRUE, FALSE);

  // add submenu to set molecule representation style
  PopupMenu *pmRep = new PopupMenu("Rendering Method");
  for(i=0; i < AtomRep::TOTAL; i++) {
    sprintf(cmdbuf, "mol modstyle %d %d %s", repindex, molindex,(AtomRepInfo[i]).name);
    pmRep->add_item((AtomRepInfo[i]).name, cmdbuf, TRUE, TRUE, atomRep->method() == i);
  }
  pm->add_submenu(pmRep, TRUE, FALSE);

  // add submenu to set molecule coloring style
  PopupMenu *pmCol = new PopupMenu("Coloring Method");
  for(i=0; i < AtomColor::TOTAL; i++) {
    sprintf(cmdbuf, "mol modcolor %d %d %s", repindex, molindex, AtomColorName[i]);
    pmCol->add_item(AtomColorName[i], cmdbuf, TRUE, TRUE, atomColor->method() == i);
  }
  pm->add_submenu(pmCol, TRUE, FALSE);

  // add commands to generally affect this representation
  sprintf(cmdbuf, "mol delrep %d %d", repindex, molindex);
  pm->add_item("Delete Rep.", cmdbuf, TRUE, FALSE);

  // return the new menu
  return pm;
}


// prepare for drawing ... do any updates needed right before draw.
void DrawMolItem::prepare(DisplayDevice *) {

  // see if the frame has changed
  if(mol->has_frame_changed()) {
    needNewFrame = TRUE;
    needRegenerate |= FORCE_REGEN;
  }
  
  create_cmdlist();
}


// change the atom coloring method.  Return success.
int DrawMolItem::change_color(AtomColor *ac) {
  if(ac) {
    *atomColor = *ac;
    needRegenerate |= COL_REGEN;
    return TRUE;
  } else
    return FALSE;
}


// change the atom rep method.  Return success.
int DrawMolItem::change_rep(AtomRep *ar) {
  if(ar) {
     *atomRep = *ar;
     needRegenerate |= REP_REGEN;
     return TRUE;
  } else {
     return FALSE;
  }
}


// change the atom selection method.  Return success.
int DrawMolItem::change_sel(AtomSel *as) {
  if(as) {
    *atomSel = *as;
    needRegenerate |= SEL_REGEN;
    return TRUE;
  } else
    return FALSE;
}

// force a recalculation of everything
void DrawMolItem::force_recalc(RegenChoices r) {
  needRegenerate |= r;
}

// return whether the Nth atom is displayed.
int DrawMolItem::atom_displayed(int n) {
  return (atomSel != NULL && mol != NULL && n >= 0 && n < mol->nAtoms &&
  		 atomSel->on[n]);
}


//////////////////////////////// drawing rep routines
void DrawMolItem::draw_lines(float *framepos) {
  float pos2[3], *fp1, *fp2;
  MolAtom *a1, *a2;
  int i,j,k,a2n,bondsdrawn;
  int lastcolor = -1;
  
  MSGDEBUG(2,"  Drawing " << mol->nAtoms << " atoms with lines and points...");
  MSGDEBUG(2, sendmsg);

  // set up general drawing characteristics for this drawing method
  // turn off material characteristics, set line style

  sprintf (commentBuffer, "MoleculeID: %d ReprID: %d Beginning Lines",
	   mol->id(), repNumber);
  cmdCommentX.putdata(commentBuffer, this);
  cmdMaterials.putdata(FALSE,this);
  cmdLineType.putdata(SOLIDLINE, this);
  cmdLineWidth.putdata(atomRep->line_thickness(), this);

  // check all atoms if they are displayed, and if so draw bonds as lines
  for(i=0; i < mol->nAtoms; i++) {
    // for each atom, draw half-bond to other displayed atoms
    if(atomSel->on[i]) {
      a1 = mol->atom(i);
      bondsdrawn = 0;

      // set color
      if (lastcolor != atomColor -> color[i]) {
	lastcolor = atomColor -> color[i];
	cmdColorIndex.putdata(lastcolor, this);
      }
      fp1 = framepos + 3*i;
      
      // draw half-bond to each bonded, displayed partner
      // NOTE: bond mid-points are calculated TWICE, once for each atom.
      // Why? So each atom is drawn entirely, including all it's half-bonds,
      // after a single set-color command, instead of having a set-color
      // command for each bond.  This reduces total drawing primitives
      // considerably, at the cost of extra calculation here.
      for(j=0; j < a1->bonds; j++) {
        a2n = a1->bondTo[j];
        if(atomSel->on[a2n]) {	// bonded atom displayed?
	  fp2 = framepos + 3*a2n;
          a2 = mol->atom(a2n);
	  for(k=0; k < 3; k++)		// calc midpoint
	    pos2[k] = 0.5 * ( *(fp1 + k) + *(fp2 + k) );
	  cmdLine.putdata(fp1, pos2, this);	// draw line
	  bondsdrawn++;			// increment counter of bonds to atom i
	}
      }
      
      // if this atom is all alone (no bonds drawn), make an 'x'
      if(!bondsdrawn) {
        cmdTextPosIndex.putdata(3*i, this);
	cmdTextX.put(this);
      }

      // indicate this atom can be picked
      pickPointIndex.putdata(3*i, i, this);
    }
  }
}


// draw all atoms just as spheres of vdw radius
void DrawMolItem::draw_just_spheres(void) {
  float radscale;
  int i;
  int lastcolor = -1;

  radscale = atomRep->sphere_rad();
  // check all atoms if they are displayed, and if so draw a sphere
  for(i=0; i < mol->nAtoms; i++) {
    if(atomSel->on[i]) {
      if (radscale > 0) {  // don't draw zero scaled spheres
	// set color
	if (lastcolor != atomColor -> color[i]) {
	  lastcolor = atomColor -> color[i];
	  cmdColorIndex.putdata(lastcolor, this);
	}
      
	// draw sphere
	cmdSphereIndex.putdata(3*i, (mol->atom(i))->radius() * radscale, this);
      }

      // indicate this atom can be picked
      pickPointIndex.putdata(3*i, i, this);
    }
  }
}


// draws each atom as a dot
void DrawMolItem::draw_points(float *framepos) {
   DispCmdPoint cmdPoint;
   sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning Points",
       mol->id(), repNumber);
   cmdCommentX.putdata(commentBuffer, this);
   cmdMaterials.putdata(FALSE,this);
   int lastcolor = -1;

   // check all atoms if they are displayed, and if so, plot a point
   for (int i=mol->nAtoms-1; i>=0; i--) {
      if (atomSel->on[i]) {
	 // set color
	 if (lastcolor != atomColor -> color[i]) {
	   lastcolor = atomColor -> color[i];
	   cmdColorIndex.putdata(lastcolor, this);
	 }

	 // draw the point
	 cmdPoint.putdata(framepos + 3*i, this);

	 // and make this pickable
	 pickPointIndex.putdata(3*i, i, this);
      }
   }
}

// draws molecule as solid, lighted VDW spheres
void DrawMolItem::draw_vdw(void) {
  MSGDEBUG(2,"  Drawing " << mol->nAtoms << " atoms as solid VDW spheres...");
  MSGDEBUG(2, sendmsg);
  sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning VDW",
	 mol->id(), repNumber);
  cmdCommentX.putdata(commentBuffer, this);
  // set up general drawing characteristics for this drawing method
  // turn on material characteristics
  cmdMaterials.putdata(TRUE,this);

  // set sphere mode and resolution
  cmdSphres.putdata(atomRep->sphere_res(), this);
  cmdSphtype.putdata(SOLIDSPHERE, this);
  
  // draw all atoms as spheres
  draw_just_spheres();

}

// draws molecule as dotted VDW spheres
void DrawMolItem::draw_dotted(void) {

  MSGDEBUG(2,"  Drawing " << mol->nAtoms << " atoms as dotted VDW spheres...");
  MSGDEBUG(2,sendmsg);

  // set up general drawing characteristics for this drawing method
  // turn off material characteristics
  sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning Dotted VDW",
	 mol->id(), repNumber);
  cmdCommentX.putdata(commentBuffer, this);
  cmdMaterials.putdata(FALSE,this);

  // set sphere mode and resolution
  cmdSphres.putdata(atomRep->sphere_res(), this);
  cmdSphtype.putdata(POINTSPHERE, this);
  
  // draw all atoms as spheres
  draw_just_spheres();
}


void DrawMolItem::draw_cpk(float *framepos) {
  float pos2[3], *fp1, *fp2;
  MolAtom *a1, *a2;
  int i,j,k,a2n,bondsdrawn;
  float radscale = 0.25 * atomRep->sphere_rad();
  float brad = 0.25 * atomRep->bond_rad();
  int bres = atomRep->bond_res();
  int lastcolor = -1;

  // special degenerate cases
  int use_spheres = TRUE;
  int use_bonds = TRUE;
  if (radscale == 0) {      // spheres of zero size, draw using bonds
    use_spheres = FALSE;
  }

  sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning CPK",
	 mol->id(), repNumber);
  cmdCommentX.putdata(commentBuffer, this);

  // draw wireframe (but still with spheres -- why not use VDW?)
  if (bres <= 2 || brad < 0.01) { // draw as lines
    use_bonds = FALSE;
  }

  MSGDEBUG(2,"  Drawing " << mol->nAtoms << " atoms with CPK method ...");
  MSGDEBUG(2,sendmsg);

  // set up general drawing characteristics for this drawing method
  // turn on material characteristics
  cmdMaterials.putdata(TRUE,this);
  
  // set sphere and cylinder mode and resolution
  cmdSphres.putdata(atomRep->sphere_res(), this);
  cmdSphtype.putdata(SOLIDSPHERE, this);

  // check all atoms if they are displayed, and if so draw bonds as cylinders,
  // and the atom as a sphere, with radius = 1/4 vdw radius
  for(i=0; i < mol->nAtoms; i++) {
    // for each atom, draw half-bond to other displayed atoms
    if(atomSel->on[i]) {
      a1 = mol->atom(i);
      bondsdrawn = 0;

      // set color
      if (lastcolor != atomColor -> color[i]) {
	lastcolor = atomColor -> color[i];
	cmdColorIndex.putdata(lastcolor, this);
      }
      fp1 = framepos + 3*i;
      
      // draw a sphere for the atom, if the size is right
      if (use_spheres) {
	cmdSphereIndex.putdata(3*i, a1->radius() * radscale, this);
      }
      
      // draw half-bond to each bonded, displayed partner
      if (use_bonds) {
	for(j=0; j < a1->bonds; j++) {
	  a2n = a1->bondTo[j];
	  if(atomSel->on[a2n]) {	// bonded atom displayed?
	    fp2 = framepos + 3*a2n;
	    a2 = mol->atom(a2n);
	    for(k=0; k < 3; k++)		// calc midpoint
	      pos2[k] = 0.5 * ( *(fp1 + k) + *(fp2 + k) );
	    
	    if(i < a2n)
	      cmdCylinder.putdata(fp1, pos2, brad, bres, this);	// draw cyl
	    else
	      cmdCylinder.putdata(pos2, fp1, brad, bres, this);
	    bondsdrawn++;			// increment counter of bonds to atom i
	  }
	}
      }

      // indicate this atom can be picked
      // yes, even when there are no sphere or bonds you can still pick
      pickPointIndex.putdata(3*i, i, this);
    }
  }
}


void DrawMolItem::draw_licorice(float *framepos) {
  float pos2[3], *fp1, *fp2;
  MolAtom *a1, *a2;
  int i,j,k,a2n,bondsdrawn;
  float brad = atomRep->bond_rad();
  int bres = atomRep->bond_res();
  int lastcolor = -1;

  sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning Licorice",
	 mol->id(), repNumber);
  cmdCommentX.putdata(commentBuffer, this);

  // special case -- if there is no thinkness, call the wireframe
  if (brad == 0) {
    draw_lines(framepos);
    return;
  }

  MSGDEBUG(2,"  Drawing " << mol->nAtoms << " atoms with licorice bonds ...");
  MSGDEBUG(2,sendmsg);

  // set up general drawing characteristics for this drawing method
  // turn on material characteristics
  cmdMaterials.putdata(TRUE,this);
  
  // set sphere and cylinder mode and resolution
  cmdSphres.putdata(atomRep->sphere_res(), this);
  cmdSphtype.putdata(SOLIDSPHERE, this);

  // check all atoms if they are displayed, and if so draw bonds as cylinders,
  // and the atom as a sphere
  for(i=0; i < mol->nAtoms; i++) {
    // for each atom, draw half-bond to other displayed atoms
    if(atomSel->on[i]) {
      a1 = mol->atom(i);
      bondsdrawn = 0;

      // set color
      if (lastcolor != atomColor -> color[i]) {
	lastcolor = atomColor -> color[i];
	cmdColorIndex.putdata(lastcolor, this);
      }
      fp1 = framepos + 3*i;
      
      // draw a sphere for the atom, with same radius as bond
      cmdSphereIndex.putdata(3*i, brad, this);
      
      // draw half-bond to each bonded, displayed partner
      for(j=0; j < a1->bonds; j++) {
        a2n = a1->bondTo[j];
        if(atomSel->on[a2n]) {	// bonded atom displayed?
	  fp2 = framepos + 3*a2n;
          a2 = mol->atom(a2n);
	  for(k=0; k < 3; k++)		// calc midpoint
	    pos2[k] = 0.5 * ( *(fp1 + k) + *(fp2 + k) );
	  if(i < a2n)
	    cmdCylinder.putdata(fp1, pos2, brad, bres, this);	// draw cyl
	  else
	    cmdCylinder.putdata(pos2, fp1, brad, bres, this);
	  bondsdrawn++;			// increment counter of bonds to atom i
	}
      }

      // indicate this atom can be picked
      pickPointIndex.putdata(3*i, i, this);
    }
  }
}

//   This draws a cylindrical bond from atom 'i' to atom 'k', but
// only if both are selected.  Each end of the cylinder is colored
// appropriately.  If there are bonds connected to the ends of this
/// i-k bond, then the i-k bond is extended so that the bonds have
// no gap between them
void DrawMolItem::draw_bonds(float *framepos)
{
   MolAtom *a1, *a2;
   int g, h, i, j, k, l, m;
   float brad = atomRep->bond_rad();
   int bres = atomRep->bond_res();
   int lastcolor = -1;

   int use_cyl;
   sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning Bonds",
	  mol->id(), repNumber);
   cmdCommentX.putdata(commentBuffer, this);
   if (bres <= 2 || brad < 0.01 ) {
     // draw bonds with lines
     use_cyl = FALSE;
     cmdMaterials.putdata(FALSE, this);
     cmdLineType.putdata(SOLIDLINE, this);
     cmdLineWidth.putdata(2, this);
   } else {
     use_cyl = TRUE;
     // set up general drawing characteristics for this drawing method
     // turn on material characteristics
     cmdMaterials.putdata(TRUE,this);
   
     // set sphere and cylinder mode and resolution
     cmdSphres.putdata(atomRep->sphere_res(), this);
     cmdSphtype.putdata(SOLIDSPHERE, this);
   }

   
   for(i=0; i < mol->nAtoms; i++) {
      // find a selected atom
      if (atomSel->on[i]) {
	 a1 = mol->atom(i);
	 // of the bonded atoms with index > i,  (See the comment
	 // later on about picking 'k' if you don't want to use the
	 // '>')
	 for (j=0; j<a1->bonds; j++) {
	    int set_pickable = FALSE;
	    // find one that is turned on
	    k = a1 -> bondTo[j];
	    if (k > i && atomSel->on[k]) {
	       a2 = mol->atom(k);
 // To make things look nice, I look for a bond that extends each end
 // of the bond.  So I look for an atom 'm' connected to 'k' which
 // is on and is not 'i'.  I also look for an atom 'g' connected to
 // 'i' which is not 'k'.
 // I deliberatly choose to look only at 'on' atoms.  I think it is
 // a good idea, but I guess you are the judge
	       for (l=a2->bonds-1; l>=0; l--) {
		  m = a2 -> bondTo[l];
		  if (m != i  && atomSel->on[m]) {
		     break;
		  }
	       } // (m is good, or l==-1)

	       for (h=a1->bonds-1; h>=0; h--) {
		  g = a1 -> bondTo[h];
		  if (g != k && atomSel->on[g]) {
		     break;
		  }
	       } // g is good, or h==-1)

	       // at this point I have:
	       // g = {good or -1}, i, k, and m = {good or -1}
	       // so I draw the appropriate cylinder
	       float *p1, *p2, p3[3], *p4, *p5;
	       if (h == -1) p1 = NULL; else p1 = framepos + 3*g;
	       p2 = framepos + 3*i;
	       p4 = framepos + 3*k;
	       if (l==-1) p5 = NULL; else p5 = framepos + 3*m;
	       // find the bisector
	       add(p3, p2, p4);
	       p3[0] /= 2.0;
	       p3[1] /= 2.0;
	       p3[2] /= 2.0;
	       // now I can draw the bonds
	       if (lastcolor != atomColor -> color[i]) {
		 lastcolor = atomColor -> color[i];
		 cmdColorIndex.putdata(lastcolor, this);
	       }
	       make_connection(p1, p2, p3, p4, brad, bres, use_cyl);
	       if (lastcolor != atomColor -> color[k]) {
		 lastcolor = atomColor -> color[k];
		 cmdColorIndex.putdata(lastcolor, this);
	       }
	       make_connection(p2, p3, p4, p5, brad, bres, use_cyl);
	       // and make it pickable (once!)
	       if (!set_pickable) {
		  set_pickable = !set_pickable;
		  pickPointIndex.putdata(3*i, i, this);
		  // this is tricky.  If 'k' is on, but the only
		  // atom bonded to it is 'i', then 'k' won't be
		  // visited.  If that is the case (ie, p5 is NULL)
		  // then I have to mark 'k' as picked here instead
		  // of later
		  if (!p5) {
		     pickPointIndex.putdata(3*k, k, this);
		  }
	       }
		  
	    }  // found k atom
	 } // searching along i's bonds
      } // found i
   } // searching each atom
}



static void make_spline_Q_matrix(float q[4][3], float basis[4][4],
				 float *pts)
{
   register i, j, dim;
   for (i = 0; i<4; i++) {
      q[i][0] = q[i][1] = q[i][2] = 0;
      for (j = 0; j<4; j++) {
	 for (dim = 0; dim<3; dim++) {
	    q[i][dim] += basis[i][j]*pts[j*3+dim];
	 }
      }
   }
}

static void make_spline_interpolation(float out[3], register float w,
				      float q[4][3])
{
   out[0] = w * (w * (w * q[0][0] + q[1][0]) + q[2][0]) + q[3][0];
   out[1] = w * (w * (w * q[0][1] + q[1][1]) + q[2][1]) + q[3][1];
   out[2] = w * (w * (w * q[0][2] + q[1][2]) + q[2][2]) + q[3][2];
}


/*
static void shift_matrix(float mat[4][3], float v[3])
{
   mat[0][0] = mat[1][0];  mat[1][0] = mat[2][0];
     mat[2][0] = mat[3][0];  mat[3][0] = v[0];
   mat[0][1] = mat[1][1];  mat[1][1] = mat[2][1];
     mat[2][1] = mat[3][1];  mat[3][1] = v[1];
   mat[0][2] = mat[1][2];  mat[1][2] = mat[2][2];
     mat[2][2] = mat[3][2];  mat[3][2] = v[2];
}
*/

// given the current cylinder and the previous and next cylinders,
// find and draw the cylinder (or line) that best fits without making
// an overlap.  prev can be NULL, in which case it acts as if the
// previous cylinder was linear with the one to draw.  Ditto (but
// opposite) for next.
//
//  The problem is that if I make two cylinders touch head to tail,
// but at an angle, there is a gap between the two.  I can get
// around it by extending the cylinders appropriately.  
void DrawMolItem::make_connection(float *prev, float *start, float *end,
				  float *next, float rad, int res, int use_cyl)
{
   if (!start || !end) {
      msgErr << "Trying to make an extended cylinder with NULL end point(s)"
	     << sendmsg;
      return;
   }
   if (!use_cyl) {
      cmdLine.putdata(start, end, this);
      return;
   }
   // now for a bit of vector math
   float new_start[3], new_end[3];
   float temp[6][3];
   float length;

   if (!prev) {          // along the previous cylinder (a)
      subtract(temp[0], end, start);  // when you don't know a
   } else {
      subtract(temp[0], start, prev);
   }
   subtract(temp[1], end, start);   // this cylinder (b)
   cross_prod(temp[2], temp[0], temp[1]);  // perp to both a and b
   // does this exceed threshhold to start extension (might speed things up)
   if (dot_prod(temp[2], temp[2]) > 0.0001/rad) {
      cross_prod(temp[3], temp[0], temp[2]);  // perp to a and away from b
      normalize(temp[3]);
      cross_prod(temp[4], temp[1], temp[2]);  // perp to b and away from a
      normalize(temp[4]);
      add(temp[5], temp[3], temp[4]);     // farthest intersection of a and b
      // find the intersection pt
      temp[5][0] *= rad/2.0; temp[5][1] *= rad/2.0; temp[5][2] *= rad/2.0;
      normalize(temp[1]);                     // unit vector along b
      length = dot_prod(temp[1], temp[5]);
   } else {
      length = 0;
   }
   new_start[0] = start[0] + length * temp[1][0];  // and the new starting
   new_start[1] = start[1] + length * temp[1][1];  // point
   new_start[2] = start[2] + length * temp[1][2];

   // this is a copy of the previous except that I swapped the
   // direction of travel.  Now a points backwards along then "next" cylinder
   // and b is this cylinder.
   if (!next) {           // backwards along the previous cylinder (a)
      subtract(temp[0], start, end);  // when you don't know a
   } else {
      subtract(temp[0], end, next);
   }
   subtract(temp[1], start, end);   // this cylinder (b)
   cross_prod(temp[2], temp[0], temp[1]);  // perp to both a and b
   // does this exceed threshhold to start extension (might speed things up)
   if (dot_prod(temp[2], temp[2]) > 0.0001/rad) {
      cross_prod(temp[3], temp[0], temp[2]);  // perp to a and away from b
      normalize(temp[3]);
      cross_prod(temp[4], temp[1], temp[2]);  // perp to b and away from a
      normalize(temp[4]);
      add(temp[5], temp[3], temp[4]);     // farthest intersection of a and b
      // find the intersection pt
      temp[5][0] *= rad/2.0; temp[5][1] *= rad/2.0; temp[5][2] *= rad/2.0;
      normalize(temp[1]);                     // unit vector along b
      length = dot_prod(temp[1], temp[5]);
   } else {
      length = 0;
   }
   new_end[0] = end[0] + length * temp[1][0];  // and the new starting
   new_end[1] = end[1] + length * temp[1][1];  // point
   new_end[2] = end[2] + length * temp[1][2];
   cmdCylinder.putdata(new_start, new_end, rad, res, this);
}

// These are different ways to define a general cubic spline
// see Foley and Van Dam, et. al., Computer Graphics, p505 or so

// this one was too sharply curved for my tastes
//   float CatmullRom_basis[4][4]={{-1.0/2.0,  3.0/2.0, -3.0/2.0,  1.0/2.0},
//				 { 2.0/2.0, -5.0/2.0,  4.0/2.0, -1.0/2.0},
//				 {-1.0/2.0,  0.0/2.0,  1.0/2.0,  0.0/2.0},
//				 { 0.0/2.0,  2.0/2.0,  0.0/2.0,  0.0/2.0}};

// this define make the next basis identical to CatmullRom
//   #define SLOPE 2.0
// This deemphasizes the slope and makes things look nicer (IMHO)
#define SLOPE 1.25
   float modified_CR_basis[4][4] = {
      {-1.0/SLOPE,  -1.0/SLOPE + 2.0,  1.0/SLOPE - 2.0,  1.0/SLOPE},
      { 2.0/SLOPE,   1.0/SLOPE - 3.0,   -2.0/SLOPE+3.0, -1.0/SLOPE},
      {-1.0/SLOPE,                 0,        1.0/SLOPE,          0},
      {         0,               1.0,                0,          0}
   };
   // This doesn't describe the system very nicely as the lines don't
   // go through the control points (which are the C-alphas)
//   float Bspline_basis[4][4]={{-1.0/6.0,  3.0/6.0, -3.0/6.0,  1.0/6.0},
//			      { 3.0/6.0, -6.0/6.0,  3.0/6.0,  0.0/6.0},
//			      {-3.0/6.0,  0.0/6.0,  3.0/6.0,  0.0/6.0},
//			      { 1.0/6.0,  4.0/6.0,  1.0/6.0,  0.0/6.0}};


// this fctn does the actual work of making the spline curve from
// a set of coordinates.  There are num points, coords goes from
// -2 to num+1 and has 3 floats per coord, idx also goes from
// -2 to num+1 and has either the atom number or -1.  The rest
// specify if cylinders should be used and (if so) the radius and
// resolution
void DrawMolItem::draw_spline_curve(int num, float *coords, int *idx,
				    int use_cyl, float b_rad, int b_res)
{
   float q[4][3];
   float prev[2][3];  // the last couple points for the previous spline
   float final[7][3]; // the points of the current approzimation
   int last_loop = -10;
   int loop;
   for (loop=-1; loop<num; loop++) { // go through the array
      
      // check if I'll need to do any computations
      // this make things faster, but its a bit more of a headache
      // later on
      if (idx[loop] >=0 && atomSel->on[idx[loop]] ||
	  idx[loop+1] >=0 && atomSel->on[idx[loop+1]]) {
	 // make the Q matrix
	 make_spline_Q_matrix(q, modified_CR_basis, coords+(loop-1)*3);
	 
	 // evaluate the interpolation between atom "loop" and
	 // "loop+1".  I'll make this in 6 pieces
	 make_spline_interpolation(final[0], 0.0/6.0, q);
	 make_spline_interpolation(final[1], 1.0/6.0, q);
	 make_spline_interpolation(final[2], 2.0/6.0, q);
	 make_spline_interpolation(final[3], 3.0/6.0, q);
	 make_spline_interpolation(final[4], 4.0/6.0, q);
	 make_spline_interpolation(final[5], 5.0/6.0, q);
	 make_spline_interpolation(final[6], 6.0/6.0, q);
	 
	 // the possible values of 'on' are 0 (off), 1 (on),
	 // 2 (draw 1st half), and 3 (draw 2nd half)
	 // However, if the 1st half is drawn, the prev atom must
	 // be on, and if the 2nd half is drawn, the new one must
	 // be on.

	 // draw what I need for atom 'loop'
	 if (idx[loop] >= 0 &&          // this is a real atom
	     (atomSel->on[idx[loop]] == 1 ||    // and it is turned on
	      (atomSel->on[idx[loop]] == 3 && idx[loop+1] >= 0 &&
	       atomSel->on[idx[loop+1]]) )) {
	    
	    // here is the trickiness I was talking about
	    // I need to see if there was a computation for
	    // the first part of this atom
	    if (last_loop != loop - 1) {  // no there wasn't
	       cmdColorIndex.putdata(atomColor->color[idx[loop]], this);
	       make_connection( NULL, final[0], final[1], final[2],
				b_rad, b_res, use_cyl);
	    } else {
       // finish up drawing the previous part of this residue (I couldn't
       // do this before since I didn't know how to angle the final cylinder
       // to get a nice match up to the first of the current cylinders)
	       make_connection( prev[0], prev[1], final[0], final[1],
				b_rad, b_res, use_cyl);
	       make_connection( prev[1], final[0], final[1], final[2],
				b_rad, b_res, use_cyl);
	    }
	    make_connection(final[0], final[1], final[2], final[3],
			    b_rad, b_res, use_cyl);
	    make_connection(final[1], final[2], final[3], final[4],
			    b_rad, b_res, use_cyl);
	    // indicate this atom can be picked
	    pickPointIndex.putdata(3*idx[loop], idx[loop], this);
	 }
	 
	 // draw what I can for atom 'loop+1'
	 if (idx[loop+1] >= 0 &&
	     (atomSel->on[idx[loop+1]] == 1 ||
	      (atomSel->on[idx[loop+1]] == 2 && idx[loop] >= 0 &&
	       atomSel->on[idx[loop]])
	     )
	    ){
	    cmdColorIndex.putdata(atomColor->color[idx[loop+1]], this);
	    make_connection(final[2], final[3], final[4], final[5],
			    b_rad, b_res, use_cyl);
	    make_connection(final[3], final[4], final[5], final[6],
			    b_rad, b_res, use_cyl);
	    last_loop = loop;
	 }
	 copy(prev[0], final[4]);  // save for the
	 copy(prev[1], final[5]);  // next interation
      } /// else nothing to draw
   } // gone down the fragment
}

// draw a spline ribbon.  The math is the same as a tube. 
// However, two splines are calculated, one for coords+perp
// and the other along coords-perp.  Triangles are used to
// connect the curves.  If cylinders are used,  they are
// drawn along the spline paths.  The final result should
// look something like:
//    ooooooooooooooooooo   where the oooo are the (optional) cylinders
//    ! /| /| /| /| /| /!   the ! are a control point, in the ribbon center
//    |/ |/ |/ |/ |/ |/ |   the /| are the triangles  (size per interpolation)
//    ooooooooooooooooooo   the edges go through coords[i] +/- offset[i]

void DrawMolItem::draw_spline_ribbon(int num, float *coords,
   float *offset, int *idx, int use_cyl, float b_rad, int b_res)
{
   float q[4][3];
   // for the next 2 variables, the information for this residue
   // starts at element 2; elements 0 and 1 are copies of the last
   // two elements of the previous residue
   float pts[2][9][3];  // data for 2 edges, 7 points (+2 overlaps from
                        // the previous interpolation) and x,y,z
   
   //  contains the norms as [edge][interpolation point][x/y/z]
   float norms[2][8][3]; 
   // contains the summed norms of the sequential triangles
   // the first element (# 0) contains the info for what should have
   // been the last two triangles of the previous residue.  In other
   // words, tri_norm[0][0] contains the summed norm for the point
   // located at pts[1][0]
   float tri_norm[2][7][3]; 

   int last_loop = -10;
   int loop, i;
   float *tmp_ptr = (float *) malloc(2*(num+4) * sizeof(float) * 3);
   if (!tmp_ptr) {
      msgErr << "Cannot make a ribbon; not enough memory!" << sendmsg;
      return;
   }
   float *edge[2];

   // copy the coordinates +/- the offsets into the temp ("edge") arrays
   edge[0] = tmp_ptr + 2*3;
   memcpy(edge[0]-2*3, coords-2*3, (num+4)*sizeof(float)*3);
   edge[1] = edge[0] + (num+4)*3;
   memcpy(edge[1]-2*3, coords-2*3, (num+4)*sizeof(float)*3);
   for (int j=-2*3; j<(num+2)*3-1; j++) {
      edge[0][j] += offset[j];
      edge[1][j] -= offset[j];
   }

   // go through the data points
   for (loop=-1; loop<num; loop++) {
      // If I'm to draw anything....
      if (idx[loop] >= 0 && atomSel->on[idx[loop]] ||
	  idx[loop+1] >= 0 && atomSel->on[idx[loop+1]] ) {

	 // construct the interpolation points (into the "pts" array)
	 // remember, these are offset by two to keep some information
	 // about the previous residue in the first 2 elements
	 for (i=0; i<=1; i++) {
	    make_spline_Q_matrix(q, modified_CR_basis, edge[i]+(loop-1)*3);
	    make_spline_interpolation(pts[i][2], 0.0/6.0, q);
	    make_spline_interpolation(pts[i][3], 1.0/6.0, q);
	    make_spline_interpolation(pts[i][4], 2.0/6.0, q);
	    make_spline_interpolation(pts[i][5], 3.0/6.0, q);
	    make_spline_interpolation(pts[i][6], 4.0/6.0, q);
	    make_spline_interpolation(pts[i][7], 5.0/6.0, q);
	    make_spline_interpolation(pts[i][8], 6.0/6.0, q);
	 }
	 // make the normals for each new point.
	 for (i=2; i<8; i++) {
	    float diff1[3], diff2[3];
	    subtract(diff1, pts[1][i+1], pts[1][i+0]);
	    subtract(diff2, pts[1][i+1], pts[0][i+0]);
	    cross_prod(norms[1][i], diff1, diff2);
	    subtract(diff1, pts[0][i+1], pts[0][i+0]);
	    cross_prod(norms[0][i], diff1, diff2);
	 }
	 //   if this wasn't a continuation, I need to set the
	 // first 2 elements properly so the norms are smooth
	 if (last_loop != loop-1) {
	    copy(norms[0][0], norms[0][2]);
	    copy(norms[1][0], norms[1][2]);
	    copy(norms[0][1], norms[0][2]);
	    copy(norms[1][1], norms[1][2]);
	 }
	 // sum up the values of neighboring triangles to make
	 // a smooth normal
	 for (i=0; i<8-1; i++) {
	    for (int j=0; j<=1; j++) {
	       add(tri_norm[j][i], norms[j][i+1],     // "this" triangle
		                   norms[1-j][i+1]);  // opposite strand
	       add(tri_norm[j][i], tri_norm[j][i],
		                   norms[j][i]);      // prev on strand
	    }
	 }
	 
	 
	 // draw what I need for atom 'loop'
	 if (idx[loop] >= 0 &&          // this is a real atom
	     atomSel->on[idx[loop]]) {  // and it it turned on
	    if (last_loop != loop - 1) {  // do prev. points exist? if not
	       // then I don't know if the color was properly set, so
	       // I'll set it here
	       cmdColorIndex.putdata(atomColor->color[idx[loop]], this);
	    }
	    // draw the cylinders to finish off the last residue, if
	    // need be, and draw the ones for this half of the residue
	    // The cylinders are drawn on the top and bottom of the
	    // residues
	    if (use_cyl) { // draw top/bot edge cylinders if need be

	       // Special case the first cylinder because I want
	       // it to be a smooth continuation from the previous
	       // cylinder; assuming it exists
	       if (last_loop != loop-1) {  // continue from previous?
		  for (int i=0; i<=1; i++) {  // doesn't exist, so
		     make_connection( NULL, pts[i][2], pts[i][3],
				      pts[i][4], b_rad, b_res, use_cyl);
		  }
	       } else { // there was a previous cylinder, so be smooth
		  for (i=0; i<=1; i++) {
	  make_connection(pts[i][0], pts[i][1], pts[i][2], pts[i][3],
			  b_rad, b_res, use_cyl);
	  make_connection(pts[i][1], pts[i][2], pts[i][3], pts[i][4],
			  b_rad, b_res, use_cyl);
                  }
	       }
	       
	       // and draw the rest of the cylinders for this 1/2 residue
	       for (i=0; i<=1; i++) {
	  make_connection( pts[i][2], pts[i][3], pts[i][4], pts[i][5],
			   b_rad, b_res, use_cyl);
	  make_connection( pts[i][3], pts[i][4], pts[i][5], pts[i][6],
			   b_rad, b_res, use_cyl);
               }
	    } // drew cylinders

	    // At this point I draw the triangles that connect the cylinders
	    // and which make up the ribbon proper
	    // the funky start condition is so that I start at
	    // pts[][1] if I need to finish the previous residue, or
	    // pts[][2] if I don't
	    for (i= (last_loop == loop-1 ? 1: 2); i<5; i++) {
       cmdTriangle.putdata(pts[1][1+i], pts[1][0+i], pts[0][0+i],
			   tri_norm[1][0+i], tri_norm[1][-1+i],
			   tri_norm[0][-1+i], this);
       cmdTriangle.putdata(pts[0][1+i], pts[1][1+i], pts[0][0+i],
			   tri_norm[0][0+i], tri_norm[1][0+i],
			   tri_norm[0][-1+i], this);

            // indicate this atom can be picked
            // this spot is in the middle of the ribbon, both in length
            // and in width
	    pickPointIndex.putdata(3*idx[loop], idx[loop], this);
            }
	 }  // atom 'loop' finished
	 
	 // draw what I can for atom 'loop+1'
	 if (idx[loop+1] >= 0 &&
	     atomSel->on[idx[loop+1]]) {
	    // If this is on, then I may have to change the color,
	    // since I'm lazy, I won't check to see if I _have_ to change it
	    // but assume I do
	    cmdColorIndex.putdata(atomColor->color[idx[loop+1]], this);
	    // I can draw the first couple of cylinders, but I need knowledge
	    // of what goes on next to get a smooth fit.  Thus, I won't
	    // draw them here.
	    // I can't draw the last two cylinders.
	    if (use_cyl) {
	       for (i=0; i<=1; i++) {
		  make_connection(pts[i][4], pts[i][5], pts[i][6], pts[i][7],
				  b_rad, b_res, use_cyl);
		  make_connection(pts[i][5], pts[i][6], pts[i][7], pts[i][8],
				  b_rad, b_res, use_cyl);
	       }
	    }
	    // I can draw 3 of the four triangles, but I need
	    // the normals to do the last one properly
	    for (i= 5; i<8-1; i++) {
/*
       cmdTriangle.putdata(pts[0][0+i], pts[1][0+i], pts[1][1+i],
			   tri_norm[0][-1+i], tri_norm[1][-1+i],
			   tri_norm[1][0+i], this);
       cmdTriangle.putdata(pts[0][0+i], pts[1][1+i], pts[0][1+i],
			   tri_norm[0][-1+i], tri_norm[1][0+i],
			   tri_norm[0][0+i], this);
*/
       cmdTriangle.putdata(pts[1][1+i], pts[1][0+i], pts[0][0+i],
			   tri_norm[1][0+i], tri_norm[1][-1+i],
			   tri_norm[0][-1+i], this);
       cmdTriangle.putdata(pts[0][1+i], pts[1][1+i], pts[0][0+i],
			   tri_norm[0][0+i], tri_norm[1][0+i],
			   tri_norm[0][-1+i], this);
            }
	    last_loop = loop;
	 }  // atom 'loop+1' finished
	 
	 // save information for the next loop through, this is needed
	 // so that the cylinders will be smooth and that the normals
	 // will look nice
	 for (i=0; i<=1; i++) {
	    copy(pts[i][0], pts[i][6]);  // remember, because of the spline,
	    copy(pts[i][1], pts[i][7]);  // loop pts[][8] is loop+1 pts[][2]
	    copy(norms[i][0], norms[i][6]);
	    copy(norms[i][1], norms[i][7]);
	 }
      } /// else nothing to draw
   } // gone down the fragment

   free(tmp_ptr);
}

// draw a cubic spline (in this case, using a modified Catmull-Rom basis)
// through the C alpha of the protein and the P of the nucleic acids
void DrawMolItem::draw_tube(float *framepos) {
   if (!mol->is_current()) {
      return;
   }
   sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning Tube",
	  mol->id(), repNumber);
   cmdCommentX.putdata(commentBuffer, this);

   float *pos;
   int atomnum, loop;
   float *real_coords = NULL, *coords;
   int *real_idx = NULL, *idx;

   // find out if I'm using lines or cylinders
   float b_rad;
   int b_res;
   if (drawmolitem_tube_radius_hack) {
     // this is a special case for drawing coils in the secondary
     // structure routines.
     b_rad = atomRep->bond_rad() / 7.0;
     b_res = int (float(atomRep->bond_res()) * 2.0 / 3.0 + 0.5);
   } else {
     b_rad = atomRep->bond_rad();
     b_res = atomRep->bond_res();
   }
   int use_cyl = FALSE;
   if (b_res <= 2 || b_rad < 0.01) { // then going to do lines
      cmdMaterials.putdata(FALSE,this);
      cmdLineType.putdata(SOLIDLINE, this);
//      cmdLineWidth.putdata(atomRep->line_thickness(), this);
      cmdLineWidth.putdata(2, this);
   } else {
      use_cyl = TRUE;
      cmdMaterials.putdata(TRUE, this);
   }
   int num, frag;

 // if there are proteins
 if (mol->pfragList.num() != 0) {
   // go through each protein fragment
   for ( frag=0; frag<mol->pfragList.num(); frag++) {
      num = mol->pfragList[frag]->num(); // number of residues
      if (num < 2) {
	continue;
      }
      if (real_coords) {
	 free(real_coords);   // to store a copy of the coords
         free(real_idx);      // and atom indicies
      }
      
      // get memory, and include space for the extra boundary elements
      // okay, this is a memory hog, sue me.
      // This allocates space for all the coordinates as well as a
      // copy of the first coordinate in -1 and -2, and similar
      // duplicates of the final coordinate.  This make the spline
      // code a lot simpler
      //    For the coordinates
      real_coords = (float *) malloc((num+4)*sizeof(float)*3);
      if (!real_coords) {
	 msgErr << "Not enough memory to make a spline" << sendmsg;
	 continue;
      }
      //    For the mapping from index in the frag list ("loop") to
      //    the offset for the give C-alpha in the molecule
      real_idx = (int *) malloc((num+4) * sizeof(int));
      if (!real_idx) {
	 msgErr << "Not enough memory to make a spline." << sendmsg;
	 free(real_coords);
	 real_coords = NULL;
	 continue;
      }
      coords = real_coords + 2*3;
      idx = real_idx + 2;

      // initialize the beginning coordinates to the first carbon-alpha
      int res = (*mol->pfragList[frag])[0]; // pfragList has residue numbers
      atomnum = mol->find_atom_in_residue("CA", res);   // get the atom index
      if (atomnum < 0) {
	 msgErr << "Cannot find the first CA of a protein!!!!!!" << sendmsg;
	 continue;
      }
      pos = framepos + 3*atomnum;     // put the first coords into
      copy(coords-6, pos);            // the initial control pts
      copy(coords-3, pos);
      idx[-2] = -1;
      idx[-1] = -1;

      // go through and set the normal spline constraints
      for (loop=0; loop<num; loop++) {   // for each residue
	 res = (*mol->pfragList[frag])[loop];
	 // find the CA and get and save the coordinates.  If for some
	 // wierd-o reason it doesn't exist, use the previous coordinates
	 atomnum = mol->find_atom_in_residue("CA", res); 
	 if (atomnum >=0) {
	   pos = framepos + 3*atomnum;
	 }
	 copy(coords+loop*3, pos); //  the coordinates
	 idx[loop] = atomnum;      // and the index
      }

      // initialize the final control points to the last element
      copy(coords+3*num, pos);
      copy(coords+3*(num+1), pos);
      idx[num]   = -1;
      idx[num+1] = -1;

      // so I have the C-alpha coordinate.  Now do the spline stuff
      // the loop starts from -1 so it can init things like the color,
      // etc.
      draw_spline_curve(num, coords, idx, use_cyl, b_rad, b_res);
      
   } // going down pfragList
 } else {
   //////////////////////////////////////////////////////////////////
   // no pfrags, SPECIAL CASE for those molecules with only CAs
   // Check for sequential records with a CA
   // and see if record(i).name == CA and record(i-1).name == CA
   //        and record(i).resid = record(i-1).resid + 1
   ResizeArray<float> x, y, z;
   ResizeArray<int> ind;
   int num = mol -> nAtoms;
   int ca_num = mol -> atomNames.typecode("CA");
   int last_resid = -10000;
   int resid;
   MolAtom *atm;
   for (int i=0; i<=num; i++) {
     if (i != num) {
       atm = mol -> atom(i);
       resid = atoi(atm -> residstr);
     } else {           // force drawing of the last atoms
       resid = -1000;
     }
     // find a sequential CA
     if (atm -> nameindex == ca_num  &&
	 resid == last_resid + 1) {
       // It is sequential, so save its info on the ResizeArrays
       x.append(framepos[3*i+0]);
       y.append(framepos[3*i+1]);
       z.append(framepos[3*i+2]);
       ind.append(i);
     } else {
       if (x.num()) {  // were there any points?
	 // draw the spline based on the ResizeArray info
	 // that means constructing some temp. info
	 if (real_coords) {
	   free(real_coords);
	   free(real_idx);
	 }
	 real_coords = (float *) malloc((x.num()+4) * sizeof(float)*3);
	 real_idx = (int *) malloc( (x.num()+4) * sizeof(int));
	 coords = real_coords + 2*3;
	 idx = real_idx + 2;
	 // move the Resize arrays to the "real" arrays
	 for (int j=0; j<x.num(); j++) {
	   coords[3*j+0] = x[j];
	   coords[3*j+1] = y[j];
	   coords[3*j+2] = z[j];
	   idx[j] = ind[j];
	 }
	 // set the header and trailer info
	 copy(coords-6, coords);
	 copy(coords-3, coords);
	 idx[-2] = idx[-1] = -1;
	 copy(coords+3*x.num(), coords+3*(x.num()-1));
	 copy(coords+3*(x.num()+1), coords+3*(x.num()-1));
	 idx[x.num()] = idx[x.num()+1] = -1;

	 // render everything
	 draw_spline_curve(x.num(), coords, idx, use_cyl, b_rad, b_res);

	 // get rid of the resize info
	 x.remove(-1, -1);
	 y.remove(-1, -1);
	 z.remove(-1, -1);
	 ind.remove(-1, -1);
       } // x.num() != 0
       // if this is a CA, save it
       if (atm -> nameindex == ca_num && i != num) {
	 x.append(framepos[3*i+0]);
	 y.append(framepos[3*i+1]);
	 z.append(framepos[3*i+2]);
	 ind.append(i);
       }
     }  // found sequential CAs

     // get ready for the next iteration
     last_resid = resid;

   } // loop over atoms
   //////////////////////////////////////////////////////////////////
 }
 ///////////////// END OF PROTEINS ///////////////////////////////

 if (mol -> nfragList.num()) {
   // now do the same with the nucleic acid backbone
   // however, in this case I use the P as my control points
   // and if there is no P I use the C5'
   for ( frag = 0; frag< mol->nfragList.num(); frag++) {
      num = mol->nfragList[frag]->num();
      if (real_coords){
	 free(real_coords);
	 free(real_idx);
      }
      // get memory
      real_coords = (float *) malloc((num+4) * sizeof(float)*3);
      if (!real_coords) {
	 msgErr << "Not enough memory to make a nucleic acid spline"
		<< sendmsg;
	 continue;
      }
      real_idx = (int *) malloc( (num+4) * sizeof(int));
      if (!real_idx) {
	 msgErr << "Not enough memory to make a nucleic acid spline."
		<< sendmsg;
	 free(real_coords);
	 real_coords = NULL;
	 continue;
      }
      coords = real_coords + 2*3;
      idx = real_idx + 2;
      int res = (*mol->nfragList[frag])[0]; // get first residue number

      // look for a phosphate.  If it isn't there, look for the C5'
      atomnum = mol->find_atom_in_residue("P", res);
      if (atomnum < 0) {  // no phosphate, so use C5'
	 atomnum = mol->find_atom_in_residue("C5'", res);
	 if (atomnum < 0) {
	    atomnum = mol->find_atom_in_residue("C5*", res);
	 }
      }
      if (atomnum < 0) {
	 msgErr << "Found a nucleic acid that doesn't have P or C5'!!!"
		<< "  Can't render it as tube." << sendmsg;
	 continue;
      }
      
      // initialize
      pos = framepos + 3*atomnum;
      copy(coords-6, pos);
      copy(coords-3, pos);
      idx[-2] = idx[-1] = -1;
      
      // now set up the rest of the coordinates
      for (loop=0; loop<num; loop++) {  // for each residue
	 res = (*mol->nfragList[frag])[loop];
	 atomnum = mol->find_atom_in_residue("P", res); // find the phosphate
	 if (atomnum >=0 ) {  // P was found
	    pos = framepos + 3*atomnum;
	 } else {
	    atomnum = mol->find_atom_in_residue("C5'", res); // or the C5'
	    if (atomnum < 0) {
	       atomnum = mol->find_atom_in_residue("C5*", res);
	    }
	    if (atomnum >= 0) {  // C5' found
	       pos = framepos + 3*atomnum;
	    }
	 }
	 // coord found, so save it
	 copy(coords+loop*3, pos);
	 idx[loop] = atomnum;
      }
      // special case if one residue so I get part of a curve
      // (I wonder if this will work; I won't test it)
      if (num == 1) {
	 atomnum = mol->find_atom_in_residue("C3'", res);
	 if (atomnum < 0) atomnum = mol->find_atom_in_residue("C3*", res);
	 if (atomnum >= 0) {
	    pos = framepos + 3*atomnum;
	 }
      }
      
      // and finish up the data by copying the last element a couple time
      copy(coords+3*num, pos);
      copy(coords+3*num+3, pos);
      idx[num] = idx[num+1] = -1;

      // render everything
      draw_spline_curve(num, coords, idx, use_cyl, b_rad, b_res);
   } // going down nfragList
 } else {
   //////////////////////////////////////////////////////////////////
   ///// SPECIAL CASE FOR JUST the phosphate of nucleic acids
   // Check for sequential records with a P
   // and see if record(i).name == P and record(i-1).name == P
   //        and record(i).resid = record(i-1).resid + 1
   ResizeArray<float> x, y, z;
   ResizeArray<int> ind;
   int num = mol -> nAtoms;
   int p_num = mol -> atomNames.typecode("P");
   int last_resid = -10000;
   int resid;
   MolAtom *atm;
   for (int i=0; i<=num; i++) {
     if (i != num) {
       atm = mol -> atom(i);
       resid = atoi(atm -> residstr);
     } else {           // force drawing of the last atoms
       resid = -1000;
     }
     // find a sequential P
     if (atm -> nameindex == p_num  &&
	 resid == last_resid + 1) {
       // It is sequential, so save its info on the ResizeArrays
       x.append(framepos[3*i+0]);
       y.append(framepos[3*i+1]);
       z.append(framepos[3*i+2]);
       ind.append(i);
     } else {
       if (x.num()) {  // were there any points?
	 // draw the spline based on the ResizeArray info
	 // that means constructing some temp. info
	 if (real_coords) {
	   free(real_coords);
	   free(real_idx);
	 }
	 real_coords = (float *) malloc((x.num()+4) * sizeof(float)*3);
	 real_idx = (int *) malloc( (x.num()+4) * sizeof(int));
	 coords = real_coords + 2*3;
	 idx = real_idx + 2;
	 // move the Resize arrays to the "real" arrays
	 for (int j=0; j<x.num(); j++) {
	   coords[3*j+0] = x[j];
	   coords[3*j+1] = y[j];
	   coords[3*j+2] = z[j];
	   idx[j] = ind[j];
	 }
	 // set the header and trailer info
	 copy(coords-6, coords);
	 copy(coords-3, coords);
	 idx[-2] = idx[-1] = -1;
	 copy(coords+3*x.num(), coords+3*(x.num()-1));
	 copy(coords+3*(x.num()+1), coords+3*(x.num()-1));
	 idx[x.num()] = idx[x.num()+1] = -1;

	 // render everything
	 draw_spline_curve(x.num(), coords, idx, use_cyl, b_rad, b_res);

	 // get rid of the resize info
	 x.remove(-1, -1);
	 y.remove(-1, -1);
	 z.remove(-1, -1);
	 ind.remove(-1, -1);
       } // x.num() != 0
       // if this is a P, save it
       if (atm -> nameindex == p_num && i != num) {
	 x.append(framepos[3*i+0]);
	 y.append(framepos[3*i+1]);
	 z.append(framepos[3*i+2]);
	 ind.append(i);
       }
     }  // found sequential Ps

     // get ready for the next iteration
     last_resid = resid;

   } // loop over atoms
   

 }
   if (real_coords) {
      free(real_coords);
      free(real_idx);
   }
}

// draw ribbons along the protein backbone
// part of this method taken (with permission) from the 'ribbon.f' in the
// Raster3D package:
//  Merritt, Ethan A. and Murphy, Michael E.P. (1994).
//   "Raster3D Version 2.0, a Program for Photorealistic Molecular Graphics"
//        Acta Cryst. D50, 869-873.

// That method was based on ideas from  Carson & Bugg, J. Molec. Graphics
//   4,121-122 (1986)
void DrawMolItem::draw_ribbons(float *framepos) {
   if (!mol->is_current()) {
      return;
   }
   sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning Ribbons",
	  mol->id(), repNumber);
   cmdCommentX.putdata(commentBuffer, this);
   float *real_coords = NULL, *coords;
   float *real_perps = NULL, *perps;
   int *real_idx = NULL, *idx;
   float *capos, *last_capos;
   float *opos, *last_opos;
   int onum, canum;
   int frag, num;
   // these are the variables used in the Raster3D package
   float a[3], b[3], c[3], d[3], e[3], g[3];  

   // find out if I'm using lines or cylinders
   float b_rad = atomRep->bond_rad()/3.0;
   int b_res = atomRep->bond_res();
   int use_cyl = FALSE;
   if (b_res <= 2 || b_rad < 0.01) { // the representation will be as lines
   } else {
      use_cyl = TRUE;
   }
   cmdMaterials.putdata(TRUE, this);
   float ribbon_width = (float) atomRep->line_thickness();
   if (ribbon_width < 1.0) ribbon_width = 1.0;
   ribbon_width /= 7.0;
   
   // go through each protein and find the CA and O
   // from that, construct the offsets to use (called "perps")
   // then do two splines and connect them
   for (frag = 0; frag<mol->pfragList.num(); frag++) {
      // number of residues in this fragment
      num = mol->pfragList[frag]->num(); 
      if (num < 2) {
// commented out since it got to be a nuisance
//	msgErr << "Cannot draw a ribbon for a protein with only one element"
//		 << sendmsg;
	 continue;
      }
      if (real_coords) {
	 free(real_coords);
	 free(real_idx);
	 free(real_perps);
      }
      real_coords = (float *) malloc((num+4)*sizeof(float)*3);
      real_idx = (int *) malloc((num+4) * sizeof(int));
      real_perps = (float *) malloc((num+4) * sizeof(float)*3);
      coords = real_coords + 2*3;
      idx = real_idx + 2;
      perps = real_perps + 2*3;

      // okay, I've got space for the coordinates, the index translations,
      // and the perpendiculars, now initialize everything
      int res = (*mol->pfragList[frag])[0];
      canum = mol->find_atom_in_residue("CA", res); // get CA 
      onum = mol->find_atom_in_residue("O", res);   //     O
      if (canum < 0) {
	 msgErr << "Cannot find first CA of a protein, so it won't be drawn"
		<< sendmsg;
	 continue;
      }
      capos = framepos + 3*canum;
      opos = framepos + 3*onum;
      last_opos = opos;
      last_capos = capos;
      
      copy(coords-6, capos);
      copy(coords-3, capos);
      idx[-2] = idx[-1] = -1;

      // now go through and set the coordinates and the perps
      e[0] = e[1] = e[2] = 0.0;
      copy(g,e);
      for (int loop=0; loop<num; loop++)
      {
	 res = (*mol->pfragList[frag])[loop];
	 canum = mol->find_atom_in_residue("CA", res);
	 if (canum >= 0) {
	    capos = framepos + 3*canum;
	 }
	 onum = mol->find_atom_in_residue("O", res);
	 if (onum < 0) {
	    onum = mol->find_atom_in_residue("OT1", res);
	 }
	 if (onum < 0) {
	    msgInfo << "Someone's been messing around with the definition "
	    "of a protein!\n" << "Things are going to look messy" << sendmsg;
	    opos = last_opos;
	 } else {
	    opos = framepos + 3*onum;
	 }
	 copy(coords+loop*3, capos);
	 idx[loop] = canum;

         // now I need to figure out where the ribbon goes
	 // A is (pos(CA(res)) - pos(CA(res-1)))
         subtract(a, capos, last_capos);
	 // B is (pos(O(res-1)) - pos(CA(res-1)))
         subtract(b, last_opos, last_capos);
         cross_prod(c, a, b); // C = A x B // defines res-1's peptide plane
         cross_prod(d, c, a); // D = C x A // normal to plane and backbone
         {
            float tempf = dot_prod(d,g);   // if the normal is > 90 degrees
            if (tempf < 0) {               // from the previous one,
               b[0] = -d[0]; b[1] = -d[1]; b[2] = -d[2];  // invert the new one
            } else {
               copy(b, d);
            }
            add(e,g,b);  // average this new vector to the cumulative sum
         }               // of the previous vectors
	 if (dot_prod(e,e) > 0.001) {  // failsafe it here since
	    normalize(e);              //  normalize doesn't
	 }
         if (loop>0) {                 // compute the offset from the
            copy(perps+3*loop, e);     // normal
	    perps[3*loop+0] *= ribbon_width;
	    perps[3*loop+1] *= ribbon_width;
	    perps[3*loop+2] *= ribbon_width;
         }
         copy(g, e);            // make a cumulative sum; cute, IMHO
         last_capos = capos;
	 last_opos = opos;
      }
      // and set the final points to the last element
      copy(coords+3*num, capos);
      copy(coords+3*(num+1), capos);
      idx[num] = idx[num+1] = -1;

      // now set the first and last perps correctly
      copy(perps, perps+3);
      copy(perps-3, perps);
      copy(perps-6, perps);
      copy(perps+3*num, perps+3*num-3);
      copy(perps+3*num+3, perps+3*num);

      draw_spline_ribbon(num, coords, perps, idx, use_cyl,
			 b_rad, b_res);
   } // drew the protein fragment ribbon


///////////// now draw the nucleic acid ribbon
   
   {
      float *ppos, *last_ppos;
      float opos[3], last_opos[3];
      int pnum;
      int o1num, o2num;
      // go through each nucleic acid and find the phospate
      // then find the O1P and O2P. From those construct the perps
      // then do two splines and connect them
      for (frag = 0; frag<mol->nfragList.num(); frag++) {
	 // number of residues in this fragment
	 num = mol->nfragList[frag]->num(); 
	 if (num < 2) {
//	msgErr << "Cannot draw a ribbon for a nucleic acid with only one element"
//		 << sendmsg;
	    continue;
	 }
	 if (real_coords) {
	    free(real_coords);
	    free(real_idx);
	    free(real_perps);
	 }
	 real_coords = (float *) malloc((num+4)*sizeof(float)*3);
	 real_idx = (int *) malloc((num+4) * sizeof(int));
	 real_perps = (float *) malloc((num+4) * sizeof(float)*3);
	 coords = real_coords + 2*3;
	 idx = real_idx + 2;
	 perps = real_perps + 2*3;
	 
	 // okay, I've got space for the coordinates, the index translations,
	 // and the perpendiculars, now initialize everything
	 int res = (*mol->nfragList[frag])[0];
	 pnum = mol->find_atom_in_residue("P", res);  // get phosphate 
	 o1num = mol->find_atom_in_residue("O1P", res); //  and an oxygen
	 o2num = mol->find_atom_in_residue("O2P", res); //  and an oxygen
	 if (pnum < 0) {
	    msgErr << "Cannot find first phosphate of a nucleic acid, so it won't be drawn"
		   << sendmsg;
	    continue;
	 }
	 if (o1num  < 0 || o2num < 0) {
	    msgErr << "Cannot find both O1P and O2P of a nucleic acid, so it won't be drawn"
		   << sendmsg;
	    continue;
	 }
	 ppos = framepos + 3*pnum;
	 add(opos, framepos + 3*o1num, framepos + 3*o2num);  // along the bisector
	 copy(last_opos, opos);
	 last_ppos = ppos;
	 
	 copy(coords-6, ppos);
	 copy(coords-3, ppos);
	 idx[-2] = idx[-1] = -1;
	 
	 // now go through and set the coordinates and the perps
	 e[0] = e[1] = e[2] = 0.0;
	 copy(g,e);
	 for (int loop=0; loop<num; loop++)
	 {
	    res = (*mol->nfragList[frag])[loop];
	    pnum = mol->find_atom_in_residue("P", res);
	    if (pnum >= 0) {
	       ppos = framepos + 3*pnum;
	    }
	    o1num = mol->find_atom_in_residue("O1P", res);
	    o2num = mol->find_atom_in_residue("O2P", res);
	    if (o1num < 0 || o2num < 0) {
	       msgInfo << "Someone's been messing around with the definition "
	       "of a nucleic acid!\n" << "Things are going to look messy" << sendmsg;
	       copy(opos, last_opos);
	    } else {
	       float tmp[3];
	       subtract(tmp, framepos + 3*o1num, ppos);
	       subtract(opos, framepos + 3*o2num, ppos);
	       add(opos, tmp, opos);  // along the bisector
	    }
	    copy(coords+loop*3, ppos);
	    idx[loop] = pnum;
	    
	    // now I need to figure out where the ribbon goes
	    // A is (pos(P(res)) - pos(P(res-1)))
	    subtract(a, ppos, last_ppos);
	    // B is (pos(Obisector(res-1)) - pos(P(res-1)))
//	    subtract(b, last_opos, last_ppos);
//	    cross_prod(c, a, b); // C = A x B // defines res-1's peptide plane
	    copy(c, opos); // already have the normal to the ribbon
	    cross_prod(d, c, a); // D = C x A // normal to plane and backbone
	    {
	       float tempf = dot_prod(d,g);   // if the normal is > 90 degrees
	       if (tempf < 0) {               // from the previous one,
		  b[0] = -d[0]; b[1] = -d[1]; b[2] = -d[2];  // invert the new one
	       } else {
		  copy(b, d);
	       }
            add(e,g,b);  // average this new vector to the cumulative sum
	    }               // of the previous vectors
	    if (dot_prod(e,e) > 0.001) {  // failsafe it here since
	       normalize(e);              //  normalize doesn't
	    }
	    if (loop>0) {                 // compute the offset from the
	       copy(perps+3*loop, e);     // normal
	       perps[3*loop+0] *= ribbon_width;
	       perps[3*loop+1] *= ribbon_width;
	       perps[3*loop+2] *= ribbon_width;
	    }
	    copy(g, e);            // make a cumulative sum; cute, IMHO
	    last_ppos = ppos;
	    copy(last_opos, opos);
	 }
	 // and set the final points to the last element
	 copy(coords+3*num, ppos);
	 copy(coords+3*(num+1), ppos);
	 idx[num] = idx[num+1] = -1;
	 
	 // now set the first and last perps correctly
	 copy(perps, perps+3);
	 copy(perps-3, perps);
	 copy(perps-6, perps);
	 copy(perps+3*num, perps+3*num-3);
	 copy(perps+3*num+3, perps+3*num);
	 
	 draw_spline_ribbon(num, coords, perps, idx, use_cyl,
			    b_rad, b_res);
       // drew the nucleic acid fragment ribbon
      }
   }

///////////// now draw the rings of the base and sugar as planes 
	// ribose 
	// 	O4',C1',C2',C3'C4' or
	// 	O4*,C1*,C2*,C3*C4*
	// purines (CYT,THY,URA) 
	// 	N1,C2,N3,C4,C5,C6
	// pyrimidines (ADE,GUA) 
	//      N1,C2,N3,C4,C5,C6,N7,N8,N9
	// O4',C1' and C4' are define the ribose plane and 
	// C2' and C3' then define the pucker of the ring
   	// sugar -- base bonds
	//  	pyrimidine 	C1' to N9 
	// 	purine 		C1' to N1  
   {
      float *o4ppos, *c1ppos, *c2ppos, *c3ppos, *c4ppos;
      int o4pnum, c1pnum, c2pnum, c3pnum, c4pnum;
      float *n1pos,*c2pos,*n3pos,*c4pos,*c5pos,*c6pos,*n7pos,*c8pos,*n9pos;
      int n1num,c2num,n3num,c4num,c5num,c6num,n7num,c8num,n9num;
      float rescentra[3], rescentrb[3];
      float midptc1pc4p[3];
      
      for (frag = 0; frag<mol->nfragList.num(); frag++) {
	 // number of residues in this fragment
	 num = mol->nfragList[frag]->num(); 
	 if (real_coords) {
	    free(real_coords);
	 }
	 //
	 // 5atoms for the ribose but only 4 triangles
	 // 9atoms max for a base
	 real_coords = (float *) malloc( 14 * num * sizeof(float)*3);
	 coords = real_coords + 2*3;
	 
	 // okay, I've got space for the coordinates now go
	 for (int loop=0; loop<num; loop++){
	 // the furanose
	    int res = (*mol->nfragList[frag])[loop];

	    c1pnum = mol->find_atom_in_residue("C1'", res);
	    if (c1pnum < 0) {
	       c1pnum = mol->find_atom_in_residue("C1*", res);
	    }
	    if ( atomSel->on[c1pnum] ){ // switch drawing by C1' atom

	      o4pnum = mol->find_atom_in_residue("O4'", res);
	      if (o4pnum < 0) {
		 o4pnum = mol->find_atom_in_residue("O4*", res);
	      }
	      c2pnum = mol->find_atom_in_residue("C2'", res); 
	      if (c2pnum < 0) {
		 c2pnum = mol->find_atom_in_residue("C2*", res);
	      }
	      c3pnum = mol->find_atom_in_residue("C3'", res); 
	      if (c3pnum < 0) {
		 c3pnum = mol->find_atom_in_residue("C3*", res);
	      }
	      c4pnum = mol->find_atom_in_residue("C4'", res);
	      if (c4pnum < 0) {
		 c4pnum = mol->find_atom_in_residue("C4*", res);
	      }
	      if (o4pnum < 0 || c2pnum < 0 || c3pnum < 0 || c4pnum < 0) {
		 msgErr << "Someone has redefined the nucleic acid atom names!"
			<< sendmsg;
		 continue;
	      }

	      o4ppos = framepos + 3*o4pnum;
	      c1ppos = framepos + 3*c1pnum;
	      c2ppos = framepos + 3*c2pnum;
	      c3ppos = framepos + 3*c3pnum;
	      c4ppos = framepos + 3*c4pnum;
	 
	      midpoint(midptc1pc4p, c1ppos, c4ppos);
	 	 
	 // now display triangles 
 	    cmdColorIndex.putdata(atomColor->color[c1pnum], this);
          
	    cmdTriangle.putdata(c4ppos,c1ppos,o4ppos,this);
            cmdTriangle.putdata(c3ppos,midptc1pc4p,c4ppos,this);
            cmdTriangle.putdata(c2ppos,midptc1pc4p,c3ppos,this);
            cmdTriangle.putdata(c1ppos,midptc1pc4p,c2ppos,this);
	   }	
	// begin bases
	  rescentra[0]=rescentra[1]=rescentra[2]=0.0;
	  rescentrb[0]=rescentrb[1]=rescentrb[2]=0.0;
	  
	  // check for purine and pyrimidine specific atoms
	  n9num = mol->find_atom_in_residue("N9", res);    	
	  n1num = mol->find_atom_in_residue("N1", res);
	    // if there is a N9, then this is a pyrimidine
	  if(( n9num >= 0) && (atomSel->on[n9num])) {
	
	     	c8num = mol->find_atom_in_residue("C8", res); 
	    	n7num = mol->find_atom_in_residue("N7", res); 
		c6num = mol->find_atom_in_residue("C6", res); 
	    	c5num = mol->find_atom_in_residue("C5", res);
	    	c4num = mol->find_atom_in_residue("C4", res);
	    	n3num = mol->find_atom_in_residue("N3", res);
	    	c2num = mol->find_atom_in_residue("C2", res);
	    	n1num = mol->find_atom_in_residue("N1", res);
		if (c8num < 0 || n7num < 0 || c6num < 0 || c5num < 0 ||
		    c4num < 0 || n3num < 0 || c2num < 0 || n1num < 0) {
 msgWarn << "Tried to draw a nucleic acid residue I thought was a pyrimidine, "
	 << sendmsg;
 msgWarn << "but it doesn't have the right atom names." << sendmsg;
                    continue;
		}

		n9pos = framepos + 3*n9num;    	
	     	add(rescentra,rescentra,n9pos);
		c8pos = framepos + 3*c8num; 
	     	add(rescentra,rescentra,c8pos);
		n7pos = framepos + 3*n7num; 
	     	add(rescentra,rescentra,n7pos);

		c5pos = framepos + 3*c5num;
	     	add(rescentra,rescentra,c5pos);
		add(rescentrb,rescentrb,c5pos);
	    	c4pos = framepos + 3*c4num;
	     	add(rescentra,rescentra,c5pos);
		add(rescentrb,rescentrb,c5pos);
		
		c6pos = framepos + 3*c6num; 
	     	add(rescentrb,rescentrb,c6pos);
	    	n3pos = framepos + 3*n3num;
	     	add(rescentrb,rescentrb,n3pos);
	    	c2pos = framepos + 3*c2num;
	     	add(rescentrb,rescentrb,c2pos);
	    	n1pos = framepos + 3*n1num;
	     	add(rescentrb,rescentrb,n1pos);
		
		rescentrb[0] = rescentrb[0]/6.0;
		rescentrb[1] = rescentrb[1]/6.0;
		rescentrb[2] = rescentrb[2]/6.0;
		
		rescentra[0] = rescentra[0]/5.0;
		rescentra[1] = rescentra[1]/5.0;
		rescentra[2] = rescentra[2]/5.0;

	       // draw bond from ribose to base
	        cmdCylinder.putdata(c1ppos, n9pos, b_rad, b_res, this);
	      // now display triangles
 	 	cmdColorIndex.putdata(atomColor->color[n9num], this);

          	cmdTriangle.putdata(n1pos,rescentrb,c2pos,this);
          	cmdTriangle.putdata(c2pos,rescentrb,n3pos,this);
          	cmdTriangle.putdata(n3pos,rescentrb,c4pos,this);
          	cmdTriangle.putdata(c4pos,rescentrb,c5pos,this);
          	cmdTriangle.putdata(c5pos,rescentrb,c6pos,this);
          	cmdTriangle.putdata(c6pos,rescentrb,n1pos,this);

          	cmdTriangle.putdata(n9pos,rescentra,c8pos,this);
          	cmdTriangle.putdata(c8pos,rescentra,n7pos,this);
          	cmdTriangle.putdata(n7pos,rescentra,c5pos,this);
          	cmdTriangle.putdata(c5pos,rescentra,c4pos,this);
          	cmdTriangle.putdata(c5pos,rescentra,c4pos,this);
          	cmdTriangle.putdata(c4pos,rescentra,n9pos,this);

		}	    
	     else if (( n1num >= 0) && (atomSel->on[n1num])){
	        // residue is purine and turned on

		c6num = mol->find_atom_in_residue("C6", res); 
	    	c5num = mol->find_atom_in_residue("C5", res);
	    	c4num = mol->find_atom_in_residue("C4", res);
	    	n3num = mol->find_atom_in_residue("N3", res);
	    	c2num = mol->find_atom_in_residue("C2", res);
		if (c6num < 0 || c5num < 0 || c4num < 0 || n3num < 0 ||
		    c2num < 0) {
 msgWarn << "Tried to draw a nucleic acid residue I thought was a purine, "
	 << sendmsg;
 msgWarn << "but it doesn't have the right atom names." << sendmsg;
                    continue;
		}

		c6pos = framepos + 3*c6num; 
	     	add(rescentrb,rescentrb,c6pos);
		c5pos = framepos + 3*c5num;
		add(rescentrb,rescentrb,c5pos);
	    	c4pos = framepos + 3*c4num;
		add(rescentrb,rescentrb,c5pos);
	    	n3pos = framepos + 3*n3num;
	     	add(rescentrb,rescentrb,n3pos);
	    	c2pos = framepos + 3*c2num;
	     	add(rescentrb,rescentrb,c2pos);
	    	n1pos = framepos + 3*n1num;
	     	add(rescentrb,rescentrb,n1pos);

		rescentrb[0] = rescentrb[0]/6.0;
		rescentrb[1] = rescentrb[1]/6.0;
		rescentrb[2] = rescentrb[2]/6.0;

 	       // draw bond from ribose to base
	        cmdCylinder.putdata(c1ppos, n1pos, b_rad, b_res, this);
	 	cmdColorIndex.putdata(atomColor->color[n1num], this);

          	cmdTriangle.putdata(n1pos,rescentrb,c2pos,this);
          	cmdTriangle.putdata(c2pos,rescentrb,n3pos,this);
          	cmdTriangle.putdata(n3pos,rescentrb,c4pos,this);
          	cmdTriangle.putdata(c4pos,rescentrb,c5pos,this);
          	cmdTriangle.putdata(c5pos,rescentrb,c6pos,this);
          	cmdTriangle.putdata(c6pos,rescentrb,n1pos,this);
		}
	
	}
       
     }
  }
     if (real_coords) {
      free(real_coords);
      free(real_idx);
      free(real_perps);
   }
}


#ifdef VMDALPHA_SHAPE

void DrawMolItem::draw_alphaShape(float *framepos) {
   int count = 0;
   float radius = atomRep->sphere_rad();
   sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning Alpha Shape",
	  mol->id(), repNumber);
   cmdCommentX.putdata(commentBuffer, this);
   // The value for get_faces was determined empirically so that
   // 1 was a reasonable value and 2 is at or near the max.
   // Without the special mods, the rough range was from 0 to 100000.
   {
      float r = radius / 2.0;
      if (r > 1.0) r = 1.0;
      if (r < 0) r = 0;
      // bring everything closer to 1
      r = sqrt(sqrt(r));
      r = exp(r * 11.513 /* = ln(100001) */ ) - 1;
//      msgInfo << "get_faces radius = " << r << sendmsg;	 
      radius = r;
   }
   int *map = new int[mol->nAtoms];  // mapping from alpha shape index
                                       // to molecular index
   // the ! new option means if only the radius was changed
   // then don't recalculate the alpha shape
//   msgInfo << "alpha == " << 0+(atomRep->alpha==NULL)
//	   << "  radius " << radius << sendmsg;
   if (atomRep -> alpha == NULL ||
       needRegenerate & FORCE_REGEN ||
       needRegenerate & SEL_REGEN  ||
       (needRegenerate & REP_REGEN && atomRep->option_changed)) {
      float *x = new float[mol->nAtoms];
      float *y = new float[mol->nAtoms];
      float *z = new float[mol->nAtoms];
      float *r = new float[mol->nAtoms];
                                    
      // check all atoms if they are displayed, and if so, get
      // the data for it
      for (int i=0; i<mol->nAtoms; i++) {
	 if (atomSel->on[i]) {
	    map[count] = i;
	    x[count] = framepos[3*i + 0];
	    y[count] = framepos[3*i + 1];
	    z[count] = framepos[3*i + 2];
	    r[count] = mol->atom(i)->radius();
	    count ++;
	 }
      }
      atomRep->set_alpha(new AlphaShape(radius, count, x, y, z, r));
      if (!atomRep->alpha->okay()) {
	 atomRep->delete_alpha();
      }
      delete [] x;
      delete [] y;
      delete [] z;
      delete [] r;
   } else { // just get the mapping
      for (int i=0; i<mol->nAtoms; i++) {
	 if (atomSel->on[i]) {
	    map[count++] = i;
	 }
      }
      // and recalculate based on previous data
      atomRep->alpha->get_faces(radius);
      if (!atomRep->alpha->okay()) {
	 atomRep->delete_alpha();
      }
   }
   atomRep -> option_changed = FALSE;

   // if there were any selected atoms for an alpha
   if (atomRep->alpha && count > 0) {  
      cmdMaterials.putdata(TRUE,this);
      // find out which verticies have been selected so I can do
      // labels, if needed
      int *vrtx = new int[mol->nAtoms];
      for (int i=0; i<mol->nAtoms; i++) {
	 vrtx[i]=0;
      }
      // play funny with the triangle - break it up into three smaller
      // quadrilaterals so that each is colored by the closest corner
      float middle[3];
      float sides[3*3];
      float *ptr1, *ptr2, *ptr3;
      int a;
      for (i=0; i<atomRep->alpha->num_triangles; i++) {
	 vrtx[a = map[atomRep->alpha->triangles[i][0]-1]] = 1;
	 ptr1 = framepos + 3*a;
	 vrtx[a = map[atomRep->alpha->triangles[i][1]-1]] = 1;
	 ptr2 = framepos + 3*a;
	 vrtx[a = map[atomRep->alpha->triangles[i][2]-1]] = 1;
	 ptr3 =  framepos + 3*a;

	 add(middle, ptr1, ptr2);
	 copy(sides, middle);
	 add(middle, middle, ptr3);
	 middle[0]/=3; middle[1]/=3; middle[2]/=3;
	 add(sides+3, ptr2, ptr3);
	 add(sides+6, ptr3, ptr1);
	 for(int j=0; j<9; j++) {
	    sides[j]/=2;
	 }
	 
	 cmdColorIndex.putdata(atomColor->
		     color[map[atomRep->alpha->triangles[i][0]-1]], this);
	 cmdTriangle.putdata(ptr1, sides, middle, this);
	 cmdTriangle.putdata(sides+6, ptr1, middle, this);

	 cmdColorIndex.putdata(atomColor->
		     color[map[atomRep->alpha->triangles[i][1]-1]], this);
	 cmdTriangle.putdata(ptr2, sides+3, middle, this);
	 cmdTriangle.putdata(sides, ptr2, middle, this);

	 cmdColorIndex.putdata(atomColor->
		     color[map[atomRep->alpha->triangles[i][2]-1]], this);
	 cmdTriangle.putdata(ptr3, sides+6, middle, this);
	 cmdTriangle.putdata(sides+3, ptr3, middle, this);
      }
      // also draw the edges
      float *f1, *f2;
      for (i=0; i<atomRep->alpha->num_edges; i++) {
	 cmdColorIndex.putdata(atomColor->
			       color[map[atomRep->alpha->edges[i][0]-1]],
			       this);
	 add(middle, f1 = (framepos + 3*map[atomRep->alpha->edges[i][0]-1]),
	             f2 = (framepos + 3*map[atomRep->alpha->edges[i][1]-1]));
	 middle[0]/=2; middle[1]/=2; middle[2]/=2;
	 cmdLine.putdata(f1, middle, this);
	 cmdLine.putdata(middle, f2, this);
      }
      // finally, display any selected atoms
      for (i=0; i<mol->nAtoms; i++) {
	 if (vrtx[i]) {
	    pickPointIndex.putdata(3*i, i, this);
	 }
      }
      delete vrtx;
   }
   delete [] map;
   msgInfo << "Alpha shape drawn." << sendmsg;
}
#endif /* VMDALPHA_SHAPE */

//   This draws a hydrogen bond from atom 'i' to the hydrogen of atom 'j', only
//  if both are selected and meet the hbond distance and/or angle criteria
//  distance is measured from heavy atom to heavy atom
//  the angle is from heavy atom to hydrogen to heavy atom
//  coloring is by the donor atom color specification
void DrawMolItem::draw_hbonds(float *framepos)
{
   MolAtom *a1, *a2;
   int i, j, k;
   float *coor1,*coor2,*hydrogen; 
   float  donortoH[3],Htoacceptor[3];
   float maxangle = atomRep->sphere_rad();
   float maxdist = atomRep->bond_rad();
   int thickness  = atomRep->line_thickness();
   maxdist *= maxdist;
   
   sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning Hbonds",
	  mol->id(), repNumber);
   cmdCommentX.putdata(commentBuffer, this);
   
   // set up general drawing characteristics for this drawing method
      cmdMaterials.putdata(FALSE,this);
      cmdLineType.putdata(DASHEDLINE, this);
      cmdLineWidth.putdata(thickness, this);
   
      
   // loop over all SELECTED atoms, inspite of the fact that non-hbonding
   //  atoms may actually be selected,searching for 
   // any pair that are closer than distance, not bonded,
   // and have a hydrogen that forms an acceptable angle.. all said 
   // and done it works pretty well only catching C-H---O hbonds 
   // only when the distance and angle are set rather unrealistic
   for(i=0; i < mol->nAtoms; i++) {
      if (atomSel->on[i] && (strncmp("H",mol->atom(i)->namestr,1)!=0) ) {
        a1 = mol->atom(i);
	coor1 = framepos + 3*i ;
	for (j=i+1; j< mol->nAtoms; j++) {
	  a2 = mol->atom(j);
	  if (atomSel->on[j] && !(a2->bonded(i)) && 	
	      (strncmp("H",mol->atom(j)->namestr,1)!=0) ) {
	    coor2 = framepos + 3*j;
	    if (distance2(coor1,coor2) < maxdist ) {
	      for(k=0; k < a2->bonds; k++){
	        if(strncmp("H",mol->atom(a2->bondTo[k])->namestr,1)==0){
	          hydrogen = framepos + 3*a2->bondTo[k];
		  subtract(donortoH,hydrogen,coor2);
	          subtract(Htoacceptor,coor1,hydrogen);
	          if (angle(donortoH, Htoacceptor)  < maxangle ) {
		    cmdColorIndex.putdata(atomColor->color[j], this);
	            cmdLine.putdata(coor1,hydrogen, this); // draw line
	          }
		}
	      }
	      for(k=0; k < a1->bonds; k++){
	        if(strncmp("H",mol->atom(a1->bondTo[k])->namestr,1)==0){
	          hydrogen = framepos + 3*a1->bondTo[k];
		  subtract(donortoH,hydrogen,coor1);
	          subtract(Htoacceptor,coor2,hydrogen);
	          if (angle(donortoH, Htoacceptor)  < maxangle ) {
		    cmdColorIndex.putdata(atomColor->color[i], this);
	            cmdLine.putdata(hydrogen,coor2, this); // draw line
	          }
		}
	      }
	    }
	  } 
	}
      }
   }
}


// find x = a*i + b where i = 0..n-1
static void least_squares(int n, const float *x, float *a, float *b) {
  float sum = 0;
  int i;
  for (i=0; i<n; i++) {    // find the sum of x
    sum += x[i];
  }
  float d = (float(n)-1.0) / 2.0;
  float t, sum_t2 = 0.0;
  *a = 0;
  for (i=0; i<n; i++) {
    t = (i - d);
    sum_t2 += t*t;
    *a += t*x[i];
  }
  *a /= sum_t2;
  *b = ( sum/float(n) - d*(*a));
}

// given the info about the helix (length, color, etc.), draw it
// This computes a best fit line long the given points.  For each "on"
// atom, a cylinder is drawn from 1/2 a residue before to 1/2 a residue
// beyond, in the right color.
void DrawMolItem::draw_alpha_helix_cylinders(ResizeArray<float> &x,
       ResizeArray<float> &y, ResizeArray<float> &z, ResizeArray<int> &atom_on,
       int *color, float bond_rad, int bond_res,
       float *start_coord, float *end_coord)
{
  // compute the best fit line for each direction
  float a[3], b[3];
  int num = x.num();
  least_squares(num, &(x[0]), a+0, b+0);
  least_squares(num, &(y[0]), a+1, b+1);
  least_squares(num, &(z[0]), a+2, b+2);
  
  // draw the cylinder(s)
  float start[3], end[3];
  
  start[0] = a[0] * (-.5) + b[0];
  start[1] = a[1] * (-.5) + b[1];
  start[2] = a[2] * (-.5) + b[2];
  copy(start_coord, start);
  for (int i=0; i<x.num(); i++) {
    end[0] = a[0] * (float(i)+0.5) + b[0];
    end[1] = a[1] * (float(i)+0.5) + b[1];
    end[2] = a[2] * (float(i)+0.5) + b[2];
    if (atom_on[i] >= 0) {
      pickPointIndex.putdata(3*atom_on[i], atom_on[i], this);
      cmdColorIndex.putdata(color[atom_on[i]], this);
      if (bond_res <= 2 || bond_rad < 0.01) {
	// the representation will be as lines
	cmdLine.putdata(start, end, this);
      } else {
	// draw the cylinder with closed ends
	cmdCylinder.putdata(start, end, bond_rad, bond_res, 1, this);
      }
    }
    memcpy(start, end, 3*sizeof(float));
  }
  copy(end_coord, start);

  // reset for the next round
  x.remove(-1, -1);
  y.remove(-1, -1);
  z.remove(-1, -1);
  atom_on.remove(-1, -1);
}

#define SCALEADD(vec, offset, byscale) {   \
    vec[0] += (*(offset  ))*byscale;       \
    vec[1] += (*(offset+1))*byscale;       \
    vec[2] += (*(offset+2))*byscale;       \
}
#define SCALESUM(vec, term1, term2, byscale) {   \
    vec[0] = (*(term1  )) + (*(term2  ))*byscale;       \
    vec[1] = (*(term1+1)) + (*(term2+1))*byscale;       \
    vec[2] = (*(term1+2)) + (*(term2+2))*byscale;       \
}
#define BETASCALE (0.2*ribbon_width)

void DrawMolItem::draw_beta_sheet(ResizeArray<float> &x,
       ResizeArray<float> &y, ResizeArray<float> &z, ResizeArray<int> &atom_on,
       int *color, float ribbon_width, float *start_coord, float *end_coord)
{
  // compute the points between the atoms, and do a linear extrapolation
  // to get the first and last point.  (I should base it on something better
  // than a linear fit ... ).
  //   I am gauranteed at least three residues for initial data
  int num = x.num();
  int i;
  float *centers = new float[3*(num+1)];
  for (i=1; i<num; i++) {   // compute the centers
    centers[3*i+0] = (x[i-1]+x[i])/2;
    centers[3*i+1] = (y[i-1]+y[i])/2;
    centers[3*i+2] = (z[i-1]+z[i])/2;
  }
  // and the linear extrapolation
  centers[0] = 2*centers[3] - centers[6];
  centers[1] = 2*centers[4] - centers[7];
  centers[2] = 2*centers[5] - centers[8];

  centers[3*num+0] = 2*centers[3*num-3] - centers[3*num-6];
  centers[3*num+1] = 2*centers[3*num-2] - centers[3*num-5];
  centers[3*num+2] = 2*centers[3*num-1] - centers[3*num-4];

  // now do the normals
  float *norms = new float[3*(num+1)];  // along the width
  float *perps = new float[3*(num+1)];  // along the height
  float d[2][3]; // deltas between successive points
  d[0][0] = x[1]-x[0];  d[0][1] = y[1]-y[0];  d[0][2] = z[1]-z[0];
  for (i=1; i<num-1; i++) {
    d[1][0] = x[i+1]-x[i];
    d[1][1] = y[i+1]-y[i];
    d[1][2] = z[i+1]-z[i];
    cross_prod(norms+3*i, d[0], d[1]);
    normalize(norms+3*i);
    if (i%2) { // flip every other normal so they are aligned
      norms[3*i+0] = -norms[3*i+0];
      norms[3*i+1] = -norms[3*i+1];
      norms[3*i+2] = -norms[3*i+2];
    }
    copy(d[0], d[1]);
  }
  // for the first one and last two normals, copy the end norms.
  copy(norms, norms+3);
  copy(norms+3*num-3, norms+3*num-6);
  copy(norms+3*num  , norms+3*num-6);

  // and the perpendiculars
  for (i=0; i<num; i++) {
    subtract(d[0], centers+3*i, centers+3*i+3);
    cross_prod(perps+3*i, d[0], norms+3*i);
    normalize(perps+3*i);
  }
  copy(perps+3*num, perps+3*num-3);

  // Draw rectangular blocks for each segment
  //  upper/lower, left/right corners;  this is the beginning of the block
  float ul[2][3], ur[2][3], ll[2][3], lr[2][3];
  SCALESUM(ul[0], centers, norms, ribbon_width);
  copy    (ll[0], ul[0]);
  SCALEADD(ul[0], perps, BETASCALE);
  SCALEADD(ll[0], perps, -BETASCALE);

  SCALESUM(ur[0] , centers, norms, -ribbon_width);
  copy    (lr[0], ur[0]);
  SCALEADD(ur[0], perps, BETASCALE);
  SCALEADD(lr[0], perps, -BETASCALE);

  // save the initial and final coordinate
  copy(start_coord, centers);
  copy(end_coord, centers+3*num);

  int prev_on = 0;
  float dot;
  for (i=0; i<num-1; i++) {  // go down the list of residues
    dot = dot_prod(perps+3*i, perps+3*i+3);
    // check the amount of rotation
    if (dot < 0) { // ribbons switched orientation
      // swap normals
      perps[3*i+3] = -perps[3*i+3];
      perps[3*i+4] = -perps[3*i+4];
      perps[3*i+5] = -perps[3*i+5];
      norms[3*i+3] = -norms[3*i+3];
      norms[3*i+4] = -norms[3*i+4];
      norms[3*i+5] = -norms[3*i+5];
    }

    // calculate the corners of the end of the blocks
    SCALESUM(ul[1], centers+3*i+3, norms+3*i+3, ribbon_width);
    copy    (ll[1], ul[1]);
    SCALEADD(ul[1], perps+3*i+3, BETASCALE);
    SCALEADD(ll[1], perps+3*i+3, -BETASCALE);
    
    SCALESUM(ur[1] , centers+3*i+3, norms+3*i+3, -ribbon_width);
    copy    (lr[1], ur[1]);
    SCALEADD(ur[1], perps+3*i+3, BETASCALE);
    SCALEADD(lr[1], perps+3*i+3, -BETASCALE);


    // draw this section, if it is on
    if (atom_on[i] >= 0) {
      pickPointIndex.putdata(3*atom_on[i], atom_on[i], this);
      cmdColorIndex.putdata(color[atom_on[i]], this);

      if (ribbon_width != 0) {
	// draw 1 or 2 segments based on how much curvature there was
	if (i == 0 || dot > .9 || dot < -.9) {
	  // small flip, so I just need one segment
	  // draw the sides
	  cmdTriangle.putdata(ur[0], ul[0], ul[1],  // top
			      perps+3*i, perps+3*i, perps+3*i+3, this);
	  cmdTriangle.putdata(ur[0], ul[1], ur[1],
			      perps+3*i, perps+3*i+3, perps+3*i+3, this);
	  cmdTriangle.putdata(lr[0], ll[0], ll[1],  // bottom
			      perps+3*i, perps+3*i, perps+3*i+3, this);
	  cmdTriangle.putdata(lr[0], ll[1], lr[1],
			      perps+3*i, perps+3*i+3, perps+3*i+3, this);
	  
	  // and the other sides
	  cmdTriangle.putdata(ul[0], ll[0], ll[1], 
			      norms+3*i, norms+3*i, norms+3*i+3, this);
	  cmdTriangle.putdata(ul[0], ll[1], ul[1], 
			      norms+3*i, norms+3*i+3, norms+3*i+3, this);
	  cmdTriangle.putdata(ur[0], lr[0], lr[1], 
			      norms+3*i, norms+3*i, norms+3*i+3, this);
	  cmdTriangle.putdata(ur[0], lr[1], ur[1], 
			      norms+3*i, norms+3*i+3, norms+3*i+3, this);
	} else {
	  // big turn, so make more lines
	  // given the construction, I know the first and last residues
	  // don't twist (the perps are just copies) so I just do a simple
	  // spline along the centers
	  float centers_q[4][3];
	  make_spline_Q_matrix(centers_q, modified_CR_basis, centers + 3*i-3);
	  
	  // make the spline for the perps and norms
	  float perps_q[4][3];
	  make_spline_Q_matrix(perps_q, modified_CR_basis, perps + 3*i-3);
	  float norms_q[4][3];
	  make_spline_Q_matrix(norms_q, modified_CR_basis, norms + 3*i-3);
	  
	  // break the section into two parts, so compute the new middle
	  float new_center[3];
	  float new_perp[3];
	  float new_norm[3];
	  make_spline_interpolation(new_center, 0.5, centers_q);
	  make_spline_interpolation(new_perp  , 0.5, perps_q);
	  normalize(new_perp);  // rescale correctly
	  make_spline_interpolation(new_norm  , 0.5, norms_q);
	  normalize(new_norm);  // rescale correctly
	  
	  // and make the new corners
	  float ul2[3], ur2[3], ll2[3], lr2[3];
	  SCALESUM(ul2 , new_center, new_norm, ribbon_width);
	  copy    (ll2, ul2);
	  SCALEADD(ul2, new_perp, BETASCALE);
	  SCALEADD(ll2, new_perp, -BETASCALE);
	  
	  SCALESUM(ur2 , new_center, new_norm, -ribbon_width);
	  copy    (lr2, ur2);
	  SCALEADD(ur2, new_perp, BETASCALE);
	  SCALEADD(lr2, new_perp, -BETASCALE);
	  
	  // draw the new, intermediate blocks
	  cmdTriangle.putdata(ur[0], ul[0], ul2,  // top
			      perps+3*i, perps+3*i, new_perp, this);
	  cmdTriangle.putdata(ur[0], ul2, ur2,
			      perps+3*i, new_perp, new_perp, this);
	  cmdTriangle.putdata(lr[0], ll[0], ll2,  // bottom
			      perps+3*i, perps+3*i, new_perp, this);
	  cmdTriangle.putdata(lr[0], ll2, lr2,
			      perps+3*i, new_perp, new_perp, this);

	// and the other sides
	cmdTriangle.putdata(ul[0], ll[0], ll2, 
			    norms+3*i, norms+3*i, new_norm, this);
	cmdTriangle.putdata(ul[0], ll2, ul2, 
			    norms+3*i, new_norm, new_norm, this);
	cmdTriangle.putdata(ur[0], lr[0], lr2, 
			    norms+3*i, norms+3*i, new_norm, this);
	cmdTriangle.putdata(ur[0], lr2, ur2, 
			    norms+3*i, new_norm, new_norm, this);
	
	// get the 1/2 half
	cmdTriangle.putdata(ur2, ul2, ul[1],  // top
			    new_perp, new_perp, perps+3*i+3, this);
	cmdTriangle.putdata(ur2, ul[1], ur[1],
			    new_perp, perps+3*i+3, perps+3*i+3, this);
	cmdTriangle.putdata(lr2, ll2, ll[1],  // bottom
			    new_perp, new_perp, perps+3*i+3, this);
	cmdTriangle.putdata(lr2, ll[1], lr[1],
			    new_perp, perps+3*i+3, perps+3*i+3, this);
	
	// and the other sides
	cmdTriangle.putdata(ul2, ll2, ll[1], 
			    new_norm, new_norm, norms+3*i+3, this);
	cmdTriangle.putdata(ul2, ll[1], ul[1], 
			    new_norm, norms+3*i+3, norms+3*i+3, this);
	cmdTriangle.putdata(ur2, lr2, lr[1], 
			    new_norm, new_norm, norms+3*i+3, this);
	cmdTriangle.putdata(ur2, lr[1], ur[1], 
			    new_norm, norms+3*i+3, norms+3*i+3, this);
	}
	
	if (!prev_on) {
	  // draw the base
	  cmdTriangle.putdata(ul[0], ll[0], ur[0], this);
	  cmdTriangle.putdata(ll[0], lr[0], ur[0], this);
	  prev_on = 1;
	}
      } else {  // ribbon_width == 0
        cmdLine.putdata(centers+3*i, centers+3*i+3, this);
      }
    } else { // atom_on[i] >= 0
      // this isn't on.  Was prev?  If so, draw its base
      if (prev_on && ribbon_width != 0) {
	cmdTriangle.putdata(ul[0], ll[0], ur[0], this);
	cmdTriangle.putdata(ll[0], lr[0], ur[0], this);
	prev_on = 0;
      }
    }

    copy(ur[0], ur[1]);
    copy(ul[0], ul[1]);
    copy(lr[0], lr[1]);
    copy(ll[0], ll[1]);
  } // end of loop though centers
  
  // but if the second to last one was on, then its end won't be visible
  if (prev_on && ribbon_width != 0) {
	cmdTriangle.putdata(ul[0], ll[0], ur[0], this);
	cmdTriangle.putdata(ll[0], lr[0], ur[0], this);
  }

  // swap the normals back, if they were swapped
  if (dot < 0) {
      perps[3*i+0] = -perps[3*i+0];
      perps[3*i+1] = -perps[3*i+1];
      perps[3*i+2] = -perps[3*i+2];
      norms[3*i+0] = -norms[3*i+0];
      norms[3*i+1] = -norms[3*i+1];
      norms[3*i+2] = -norms[3*i+2];
  }

  // what remains is the last one, which will be drawn as an arrow head
  if (atom_on[num-1] >= 0) {
    pickPointIndex.putdata(3*atom_on[num-1], atom_on[num-1], this);
    // recompute the base and tip information correctly
    // the 'norms' direction is 50% longer in each direction
    norms[3*num-3] *= 1.5;
    norms[3*num-2] *= 1.5;
    norms[3*num-1] *= 1.5;

    SCALESUM(ul[0], centers+3*num-3, norms+3*num-3, ribbon_width);
    copy    (ll[0], ul[0]);
    SCALEADD(ul[0], perps+3*num-3, BETASCALE);
    SCALEADD(ll[0], perps+3*num-3, -BETASCALE);

    SCALESUM(ur[0], centers+3*num-3, norms+3*num-3, -ribbon_width);
    copy    (lr[0], ur[0]);
    SCALEADD(ur[0], perps+3*num-3, BETASCALE);
    SCALEADD(lr[0], perps+3*num-3, -BETASCALE);

    // and the tip has no norms component
    copy    (ur[1], centers+3*num);
    copy    (lr[1], ur[1]);
    SCALEADD(ur[1], perps+3*num, BETASCALE);
    SCALEADD(lr[1], perps+3*num, -BETASCALE);

    // draw the arrow
    cmdColorIndex.putdata(color[atom_on[num-1]], this);
if (ribbon_width > 0) {
    // the normals are the same (from a copy) so I don't worry about them
    cmdTriangle.putdata(ul[0], ur[1], ur[0], this);    // top
    // but that means here I _do_ have to switch the normal direction
    cmdTriangle.putdata(ll[0], lr[0], lr[1], this);    // bottom
    
    cmdTriangle.putdata(ul[0], ll[0], lr[1], this);  // sides
    cmdTriangle.putdata(ul[0], lr[1], ur[1], this);
    cmdTriangle.putdata(ur[0], lr[0], lr[1], this);
    cmdTriangle.putdata(ur[0], lr[1], ur[1], this);

    cmdTriangle.putdata(ul[0], ll[0], ur[0], this);  // and base
    cmdTriangle.putdata(ll[0], lr[0], ur[0], this);
} else {// ribbon_width == 0
    cmdLine.putdata(centers+3*num-3, centers+3*num, this);
} // special case for ribbon_width == 0
  }
  delete [] perps;
  delete [] norms;
  delete [] centers;

  // reset for the next round
  x.remove(-1, -1);
  y.remove(-1, -1);
  z.remove(-1, -1);
  atom_on.remove(-1, -1);
}


// draw based on secondary structure elements
void DrawMolItem::draw_structure(float *framepos)
{
  // Indicate that I need secondary structure information
  sprintf (commentBuffer,"MoleculeID: %d ReprID: %d Beginning Cartoon",
	 mol->id(), repNumber);
  cmdCommentX.putdata(commentBuffer, this);
  mol->need_secondary_structure(1);  // calculate 2ndary structure if need be

  //
  float bond_rad = atomRep->bond_rad();
  int bond_res = atomRep->bond_res();

  ResizeArray<float> resx; // coord of start/end of structure
  ResizeArray<float> resy;
  ResizeArray<float> resz;
  ResizeArray<int> CA_num; // CA of start/end structure
  ResizeArray<int> extra_resid; // CA of very short residues


  ResizeArray<float> x; // coords (x,y,z) of the residues
  ResizeArray<float> y;
  ResizeArray<float> z;
  // used to turn on/off unselected parts of the helix
  ResizeArray<int> atom_on;

  // go along the secondary structure of the proteins and draw
  // alpha helicies as cylinders
  int h_start = -1;
  int atom;
  int frag, res;
  if (bond_res <= 2 || bond_rad < 0.01) {  // for the helicies
    cmdLineType.putdata(SOLIDLINE, this);
    cmdLineWidth.putdata(2, this);
    cmdMaterials.putdata(FALSE, this);
  } else {
    cmdMaterials.putdata(TRUE,this);
  }
  for (frag=0; frag<mol->pfragList.num(); frag++) {
    int num = mol->pfragList[frag]->num();
    for (int resindex=0; resindex < num; resindex++) {
      res = (*mol->pfragList[frag])[resindex];
      Residue::SStruct ss = mol->residue(res)->sstruct;
      if (ss == Residue::HELIX_ALPHA || ss == Residue::HELIX_3_10 ||
	   ss == Residue::HELIX_PI) {
	
	atom = mol -> find_atom_in_residue("CA", res);
	if (atom >= 0) {
	  if (h_start == -1) {
	    // just started a helix
	    h_start = atom;
	  }
	  // add the coordinate to the arrays
	  x.append(framepos[3*atom+0]);
	  y.append(framepos[3*atom+1]);
	  z.append(framepos[3*atom+2]);
	  if (atomSel -> on[atom]) {  // atom_on contains either
	    atom_on.append(atom);       //  the atom index (if on)
	  } else {
	    atom_on.append(-1);         //  or -1 (if off)
	  }
	} else {
	  msgErr << "Missing a CA in a protein residue!!" << sendmsg;
	}

      } else if (h_start != -1) {  // Here we have an alpha helix
	// so draw it
	if (x.num() <= 1) {
	  //	  msgWarn << "Structure will not draw a helix of length 1" 
	  //		  << sendmsg;
	  // Keep track of these for the coil
	  if (x.num() == 1) {
	    extra_resid.append(atom);
	  }
	  // and clear out the arrays
	  x.remove(-1, -1);
	  y.remove(-1, -1);
	  z.remove(-1, -1);
	} else {
	  float ends[2][3];
	  CA_num.append(h_start);// keep track of the residues which start
	  CA_num.append(atom);  //  and end the given secondary structure
	  draw_alpha_helix_cylinders(x, y, z, atom_on, atomColor->color, 
				     bond_rad, bond_res, ends[0], ends[1]);
	  resx.append(ends[0][0]);  // save the start/end coordinates
	  resy.append(ends[0][1]);
	  resz.append(ends[0][2]);
	  resx.append(ends[1][0]);
	  resy.append(ends[1][1]);
	  resz.append(ends[1][2]);
	}
	h_start = -1; // reset the counter
      } // drew helix

    } // went through residues

    // but the last one could still be a helix so
    if (h_start != -1) {
      if (x.num() <= 1) {
	//	msgWarn << "Will not draw a helix of length 1" << sendmsg;
	// and clear out the arrays
	extra_resid.append(h_start);
	x.remove(-1, -1);
	y.remove(-1, -1);
	z.remove(-1, -1);
      } else {
	float ends[2][3];
	CA_num.append(h_start); // keep track of the residues which start
	CA_num.append(atom);     //  and end the given secondary structure
	draw_alpha_helix_cylinders( x, y, z, atom_on, atomColor->color,
				    bond_rad, bond_res, ends[0], ends[1]);
	resx.append(ends[0][0]);  // save the start/end coordinates
	resy.append(ends[0][1]);
	resz.append(ends[0][2]);
	resx.append(ends[1][0]);
	resy.append(ends[1][1]);
	resz.append(ends[1][2]);
      }
    } // end of final check for helix
  } // went through pfrag

  // now do beta sheets via another pass through the pfrags
  if (atomRep->line_thickness() == 0) {
    cmdLineType.putdata(SOLIDLINE, this);
    cmdLineWidth.putdata(2, this);
    cmdMaterials.putdata(FALSE, this);
  } else {
    cmdMaterials.putdata(TRUE,this);
  }
  h_start = -1;
  for (frag=0; frag<mol->pfragList.num(); frag++) {
    int num = mol->pfragList[frag]->num();
    for (int resindex=0; resindex < num; resindex++) {
      int res = (*mol->pfragList[frag])[resindex];
      Residue::SStruct ss = mol->residue(res)->sstruct;
      if (ss == Residue::BETA || ss == Residue::BRIDGE) {

	atom = mol -> find_atom_in_residue("CA", res);
	if (atom >= 0) {
	  if (h_start == -1) {
	    // just started a sheet
	    h_start = atom;
	  }
	  // add the coordinate to the arrays
	  x.append(framepos[3*atom+0]);
	  y.append(framepos[3*atom+1]);
	  z.append(framepos[3*atom+2]);
	  if (atomSel -> on[atom]) {  // atom_on contains either
	    atom_on.append(atom);       //  the atom index (if on)
	  } else {
	    atom_on.append(-1);         //  or -1 (if off)
	  }
	} else {
	  msgErr << "Missing a CA in a protein residue!!" << sendmsg;
	}
      } else if (h_start != -1) { // I have a beta sheet, so draw it
	if (x.num() <= 2) {
	  //	  msgWarn << "Structure will not draw a sheet with less than "
	  //	          << "3 residues" << sendmsg;
	  // and clear out the arrays
	  extra_resid.append(h_start);
	  if (x.num() == 2) {
	    extra_resid.append(atom);
	  }
	  x.remove(-1, -1);
	  y.remove(-1, -1);
	  z.remove(-1, -1);
	  atom_on.remove(-1, -1);
	} else {
	  float ends[2][3];
	  CA_num.append(h_start); // keep track of the residues which start
	  CA_num.append(atom);     //  and end the given secondary structure
	  draw_beta_sheet(x, y, z, atom_on, atomColor->color,
			  atomRep->line_thickness() / 5.0, ends[0], ends[1]);
	  resx.append(ends[0][0]);  // save the start/end coordinates
	  resy.append(ends[0][1]);
	  resz.append(ends[0][2]);
	  resx.append(ends[1][0]);
	  resy.append(ends[1][1]);
	  resz.append(ends[1][2]);
	}
	h_start = -1; // reset
      }
    }  // went through the residues
    // but the last one could still be a sheet, so
    if (h_start != -1) {
      if (x.num() <= 2) {
	//	  msgWarn << "Structure will not draw a sheet with less than "
	//	          << "3 residues" << sendmsg;
	extra_resid.append(h_start);
	if (x.num() == 2) {
	  extra_resid.append(atom);
	}
      } else {
	float ends[2][3];
	CA_num.append(h_start); // keep track of the residues which start
	CA_num.append(res);     //  and end the given secondary structure
	draw_beta_sheet(x, y, z, atom_on, atomColor->color,
			atomRep->line_thickness() / 5.0, ends[0], ends[1]);
	resx.append(ends[0][0]);  // save the start/end coordinates
	resy.append(ends[0][1]);
	resz.append(ends[0][2]);
	resx.append(ends[1][0]);
	resy.append(ends[1][1]);
	resz.append(ends[1][2]);
      }
    }
  } // went through pfrag

  // Finally, everything left I can draw as a spline.  The tricky part
  // is I need to get the start/end coordinates of the cylinders and
  // sheets correct, for otherwise  the tube doesn't meet with them nicely.
  // Luckily, I kept track of the start/end coordinates as I went along, so
  // I swap the coords of that array with the frampos data, replace atomSel'
  // 'on' with 'newon', and call draw_tube.  Code reuse, right?  Anyway, when
  // I return I swap everything back
  {
    // finding the atoms is easy; search for 'protein not (sheet or helix)'
    // and not those atoms in the same residue as the extra_resid list.
    AtomSel new_sel(moleculeList);
    new_sel.change("protein and not (sheet or helix)");
    new_sel.find(mol);
    // get the residues in "short" sheets or cylinders
    int i;
    for (i=0; i<extra_resid.num(); i++) {
      int res = mol->atom(extra_resid[i])->uniq_resid;
      Residue *r = mol->residue(res);
      for (int j=0; j< r->atoms.num(); j++) {
	new_sel.on[r->atoms[j]] = 1;
      }
    }

    // replace the coordinates
    float tmp_coords[3];
    int atom;
    for (i=0; i<CA_num.num(); i++) {
      atom = 3*CA_num[i];
      copy(tmp_coords, framepos+atom);
      framepos[atom+0] = resx[i];
      framepos[atom+1] = resy[i];
      framepos[atom+2] = resz[i];
      resx[i] = tmp_coords[0];
      resy[i] = tmp_coords[1];
      resz[i] = tmp_coords[2];
      // turn CA of the start/end on as well
      new_sel.on[CA_num[i]] = 2+i%2;    ////////  <<<<-----  this is "2" or "3"
      // the "2" is a special case for draw_spline_curve which will only
      // draw the first 1/2 if it is a 2 or second half if it is a 3
    }

    // replace the list of "on" atoms
    for (i=0; i<mol->nAtoms; i++) {
      if (!atomSel->on[i]) { // mask the coil against the selected atoms
	new_sel.on[i] = 0;
      }
    }
    int *temp_on = atomSel->on;
    atomSel->on = new_sel.on;

    // make the tube
    drawmolitem_tube_radius_hack = 1;
    draw_tube(framepos);
    drawmolitem_tube_radius_hack = 0;

    // and revert to the original coords
    atomSel->on = temp_on;
    for ( i=0; i<CA_num.num(); i++) {
      atom = 3*CA_num[i];
      framepos[atom+0] = resx[i];
      framepos[atom+1] = resy[i];
      framepos[atom+2] = resz[i];
    }
  }
  cmdMaterials.putdata(FALSE,this);
}

