/***************************************************************************
 *cr                                                                       
 *cr            (C) Copyright 1995 The Board of Trustees of the           
 *cr                        University of Illinois                       
 *cr                         All Rights Reserved                        
 *cr                                                                   
 ***************************************************************************/

/***************************************************************************
 * RCS INFORMATION:
 *
 *	$RCSfile: Pickable.h,v $
 *	$Author: billh $	$Locker:  $		$State: Exp $
 *	$Revision: 1.5 $	$Date: 1995/11/15 04:29:29 $
 *
 ***************************************************************************
 * DESCRIPTION:
 *
 * A Pickable object is one which contains data which may be selected by
 * using a pointer device in an open DisplayDevice.  Each Pickable registers
 * itself with a PickList object which indicates it has items in its
 * drawing list which may be picked, and which indicates it should be told
 * when something has been successfully selected.  There are one or many
 * picking modes, which are represented by derivations of the PickMode class.
 * When picking is done, it is done while in a current pick mode, which
 * controls what to do with the information.
 *
 * A Pickable must provide versions of virtual routines which tell what
 * pick modes the object is interested in.
 ***************************************************************************/
#ifndef PICKABLE_H
#define PICKABLE_H

#include "utilities.h"
class PickList;
class PickMode;
class DisplayDevice;
class PopupMenu;


class Pickable {

protected:
  // the picklist with which we registered
  PickList *pickList;

  // creates a new pick mode of a type specified by the code integer argument.
  // this is typically only used by add_pick_mode; the code is both defined
  // by and used by the derived class
  virtual PickMode *create_pick_mode(int);

public:
  // constructor: PickList with which to register.  If NULL, do nothing
  Pickable(PickList * = NULL);
  
  // destructor: unregister with the picklist
  virtual ~Pickable(void);

  //
  // public utility routines
  //
  
  // register with the given PickList.  If already registered, deregisters
  // with the old one first
  void register_with_picklist(PickList *);

  // check to see if a PickMode with the given name has been created yet
  // for the PickList.  If so, return the index of the mode.  If not,
  // this routine calls the virtual create_pick_mode with the 2nd argument,
  // which is a code for the derived class to know what kind of PickMode
  // object to create.  Returns (-1) if some sort of error occurs.
  int add_pick_mode(char *, int);

  //
  // public virtual informative routines
  //
  
  // return if we are interested in the given pick mode or not
  virtual int want_pick_mode(int);
  
  // return our list of draw commands with picking draw commands in them
  virtual void *pick_cmd_list(void);
  
  // return name of pickable
  virtual char *pick_name(void);

  // return whether the pickable object is being displayed
  virtual int pickable_on(void);

  //
  // public virtual specialized popup menu routines
  //

  // Create a popup menu, with the given name; return NULL if none available
  // subclasses should have the parent first create a menu, and then add items
  // to the end of it.  By default, this does nothing and returns NULL.
  // 1st argument is the tag of the item selected (it should correspond to
  // a tag from this object).
  virtual PopupMenu *create_popup_menu(int);

  //
  // public virtual pick action routines
  //

  // called when a pick is begun:
  //	args = display to use, obj picked, button, mode, tag, dim, pos
  // For 2D version: x & y are 0 ... 1, represent 'relative, scaled' coords.
  // For 3D version: x,y,z are transformed position of pointer
  virtual void pick_start(DisplayDevice *, Pickable *,
  				int, int, int, int, float *);
  
  // called when a pick moves:
  //	args = display to use, obj picked, button, mode, tag, dim, pos
  // For 2D version: x & y are 0 ... 1, represent 'relative, scaled' coords.
  // For 3D version: x,y,z are transformed position of pointer
  virtual void pick_move(DisplayDevice *, Pickable *,
  				int, int, int, int, float *);
  
  // called when a pick ends:
  //	args = display to use, obj picked, button, mode, tag, dim, pos
  // For 2D version: x & y are 0 ... 1, represent 'relative, scaled' coords.
  // For 3D version: x,y,z are transformed position of pointer
  virtual void pick_end(DisplayDevice *, Pickable *,
  				int, int, int, int, float *);

};

#endif

